# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# pylint: disable=protected-access
"""Tests for triage task."""

import datetime
import unittest

from clusterfuzz._internal.datastore import data_handler
from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.tests.test_libs import appengine_test_utils
from clusterfuzz._internal.tests.test_libs import helpers
from clusterfuzz._internal.tests.test_libs import test_utils
from handlers.cron import triage


@test_utils.with_cloud_emulators('datastore')
class CrashImportantTest(unittest.TestCase):
  """Tests for _is_crash_important."""

  def setUp(self):
    helpers.patch(self, [
        'clusterfuzz._internal.metrics.crash_stats.get_last_successful_hour',
        'clusterfuzz._internal.metrics.crash_stats.get',
        'clusterfuzz._internal.base.utils.utcnow',
    ])
    self.mock.utcnow.return_value = test_utils.CURRENT_TIME

  def test_is_crash_important_1(self):
    """Ensure that a reproducible testcase is important."""
    testcase = test_utils.create_generic_testcase()
    testcase.one_time_crasher_flag = False
    testcase.put()

    self.assertTrue(triage._is_crash_important(testcase))

  def test_is_crash_important_2(self):
    """Ensure that an unreproducible testcase with status Unreproducible is
    not important."""
    testcase = test_utils.create_generic_testcase()
    testcase.one_time_crasher_flag = True
    testcase.status = 'Unreproducible'
    testcase.put()

    self.assertFalse(triage._is_crash_important(testcase))

  def test_is_crash_important_3(self):
    """Ensure that an unreproducible testcase with status Duplicate is
    not important."""
    testcase = test_utils.create_generic_testcase()
    testcase.one_time_crasher_flag = True
    testcase.status = 'Duplicate'
    testcase.put()

    self.assertFalse(triage._is_crash_important(testcase))

  def test_is_crash_important_4(self):
    """If the unreproducible testcase has another reproducible testcase in
    group, then crash is not important."""
    testcase_1 = test_utils.create_generic_testcase()
    testcase_1.one_time_crasher_flag = True
    testcase_1.group_id = 1
    testcase_1.put()

    testcase_2 = test_utils.create_generic_testcase()
    testcase_2.one_time_crasher_flag = False
    testcase_2.group_id = 1
    testcase_2.put()

    self.assertFalse(triage._is_crash_important(testcase_1))

  def test_is_crash_important_5(self):
    """If we don't have any crash stats data for this unreproducible testcase,
    then we can't make judgement on crash importance, so we return result as
    False."""
    self.mock.get_last_successful_hour.return_value = None
    testcase = test_utils.create_generic_testcase()
    testcase.one_time_crasher_flag = True
    testcase.put()

    self.assertFalse(triage._is_crash_important(testcase))

  def test_is_crash_important_6(self):
    """If this unreproducible testcase is less than the total crash threshold,
    then it is not important."""
    self.mock.get_last_successful_hour.return_value = 417325
    indices = [{
        'count': 1,
        'hour': day_index
    } for day_index in range(417325, 416989, -24)]
    self.mock.get.return_value = (1, [{
        'totalCount': 14,
        'groups': [{
            'indices': indices,
            'name': 'false',
        },]
    }])
    testcase = test_utils.create_generic_testcase()
    testcase.one_time_crasher_flag = True
    testcase.put()

    self.assertFalse(triage._is_crash_important(testcase))

  def test_is_crash_important_7(self):
    """If this unreproducible testcase spiked only for a certain interval, then
    it is not important."""
    self.mock.get_last_successful_hour.return_value = 417325
    self.mock.get.return_value = (1, [{
        'totalCount':
            125,
        'groups': [{
            'indices': [{
                'count': 125,
                'hour': 417301,
            }],
            'name': 'false',
        },]
    }])
    testcase = test_utils.create_generic_testcase()
    testcase.one_time_crasher_flag = True
    testcase.put()

    self.assertFalse(triage._is_crash_important(testcase))

  def test_is_crash_important_8(self):
    """If this unreproducible testcase is crashing frequently, then it is an
    important crash."""
    self.mock.get_last_successful_hour.return_value = 417325
    indices = [{
        'count': 10,
        'hour': day_index
    } for day_index in range(417325, 416989, -24)]
    self.mock.get.return_value = (1, [{
        'totalCount': 140,
        'groups': [{
            'indices': indices,
            'name': 'false',
        },]
    }])
    testcase = test_utils.create_generic_testcase()
    testcase.one_time_crasher_flag = True
    testcase.put()

    self.assertTrue(triage._is_crash_important(testcase))

  def test_is_crash_important_9(self):
    """If this unreproducible testcase is crashing frequently, but its crash
    type is one of crash type ignores, then it is not an important crash."""
    self.mock.get_last_successful_hour.return_value = 417325
    indices = [{
        'count': 10,
        'hour': day_index
    } for day_index in range(417325, 416989, -24)]
    self.mock.get.return_value = (1, [{
        'totalCount': 140,
        'groups': [{
            'indices': indices,
            'name': 'false',
        },]
    }])
    testcase = test_utils.create_generic_testcase()
    testcase.one_time_crasher_flag = True
    testcase.put()

    for crash_type in ['Out-of-memory', 'Stack-overflow', 'Timeout']:
      testcase.crash_type = crash_type
      self.assertFalse(triage._is_crash_important(testcase))


@test_utils.with_cloud_emulators('datastore')
class CheckAndUpdateSimilarBug(unittest.TestCase):
  """Tests for _check_and_update_similar_bug."""

  def setUp(self):
    helpers.patch(self, [
        'clusterfuzz._internal.base.utils.utcnow',
    ])
    self.mock.utcnow.return_value = test_utils.CURRENT_TIME

    self.testcase = test_utils.create_generic_testcase()
    self.issue = appengine_test_utils.create_generic_issue()
    self.issue_tracker = self.issue.issue_tracker

  def test_no_other_testcase(self):
    """Tests result is false when there is no other similar testcase."""
    self.assertEqual(
        False,
        triage._check_and_update_similar_bug(self.testcase, self.issue_tracker))

  def test_similar_testcase_without_bug_information(self):
    """Tests result is false when there is a similar testcase but without an
    associated bug."""
    similar_testcase = test_utils.create_generic_testcase()  # pylint: disable=unused-variable

    self.assertEqual(
        False,
        triage._check_and_update_similar_bug(self.testcase, self.issue_tracker))

  def test_similar_testcase_get_issue_failed(self):
    """Tests result is false when there is a similar testcase with an associated
    bug but we are unable to fetch it via get_issue."""
    similar_testcase = test_utils.create_generic_testcase()
    similar_testcase.bug_information = '2'  # Non-existent.
    similar_testcase.put()

    self.assertEqual(
        False,
        triage._check_and_update_similar_bug(self.testcase, self.issue_tracker))

  def test_similar_testcase_is_reproducible_and_open(self):
    """Tests result is true when there is a similar testcase which is
    reproducible, open and has an accessible associated bug."""
    self.issue.save()

    similar_testcase = test_utils.create_generic_testcase()
    similar_testcase.one_time_crasher_flag = False
    similar_testcase.open = True
    similar_testcase.bug_information = str(self.issue.id)
    similar_testcase.put()

    self.assertEqual(
        True,
        triage._check_and_update_similar_bug(self.testcase, self.issue_tracker))

  def test_similar_testcase_reproducible_and_closed_but_issue_open_1(self):
    """Tests result is true when there is a similar testcase which is
    reproducible and fixed due to flakiness but issue is kept open. Only update
    testcase bug mapping if similar testcase is fixed longer than the grace
    period."""
    self.issue.save()

    similar_testcase = test_utils.create_generic_testcase()
    similar_testcase.one_time_crasher_flag = False
    similar_testcase.open = False
    similar_testcase.bug_information = str(self.issue.id)
    similar_testcase.put()

    self.assertEqual(
        True,
        triage._check_and_update_similar_bug(self.testcase, self.issue_tracker))
    testcase = data_handler.get_testcase_by_id(self.testcase.key.id())
    self.assertEqual(None, testcase.bug_information)
    self.assertEqual('', self.issue._monorail_issue.comment)

    similar_testcase.set_metadata(
        'closed_time',
        test_utils.CURRENT_TIME -
        datetime.timedelta(hours=data_types.MIN_ELAPSED_TIME_SINCE_FIXED + 1))
    self.assertEqual(
        True,
        triage._check_and_update_similar_bug(self.testcase, self.issue_tracker))

  def test_similar_testcase_reproducible_and_closed_but_issue_open_2(self):
    """Tests result is true when there is a similar testcase which is
    reproducible and fixed due to flakiness but issue is kept open. Don't update
    testcase bug mapping if another reproducible testcase is open and attached
    to this bug."""
    self.issue.save()

    similar_testcase_1 = test_utils.create_generic_testcase()
    similar_testcase_1.one_time_crasher_flag = False
    similar_testcase_1.open = False
    similar_testcase_1.bug_information = str(self.issue.id)
    similar_testcase_1.put()

    similar_testcase_2 = test_utils.create_generic_testcase()
    similar_testcase_2.one_time_crasher_flag = False
    similar_testcase_2.open = True
    similar_testcase_2.bug_information = str(self.issue.id)
    similar_testcase_2.put()

    self.assertEqual(
        True,
        triage._check_and_update_similar_bug(self.testcase, self.issue_tracker))
    testcase = data_handler.get_testcase_by_id(self.testcase.key.id())
    self.assertEqual(None, testcase.bug_information)
    self.assertEqual('', self.issue._monorail_issue.comment)

  def test_similar_testcase_unreproducible_but_issue_open(self):
    """Tests result is true when there is a similar testcase which is
    unreproducible but issue is kept open. Update testcase bug mapping always
    since this testcase is reproducible."""
    self.issue.save()

    similar_testcase = test_utils.create_generic_testcase()
    similar_testcase.one_time_crasher_flag = True
    similar_testcase.open = False
    similar_testcase.bug_information = str(self.issue.id)
    similar_testcase.put()

    self.assertEqual(
        True,
        triage._check_and_update_similar_bug(self.testcase, self.issue_tracker))

  def test_similar_testcase_with_issue_closed_with_ignore_label(self):
    """Tests result is true when there is a similar testcase with closed issue
    blacklisted with ignore label."""
    self.issue.status = 'WontFix'
    self.issue._monorail_issue.open = False
    self.issue.labels.add('ClusterFuzz-Ignore')
    self.issue.save()

    similar_testcase = test_utils.create_generic_testcase()
    similar_testcase.one_time_crasher_flag = False
    similar_testcase.open = False
    similar_testcase.bug_information = str(self.issue.id)
    similar_testcase.put()

    self.assertEqual(
        True,
        triage._check_and_update_similar_bug(self.testcase, self.issue_tracker))

    testcase = data_handler.get_testcase_by_id(self.testcase.key.id())
    self.assertEqual(
        'Skipping filing a bug since similar testcase (2) in issue (1) '
        'is blacklisted with ClusterFuzz-Ignore label.',
        testcase.get_metadata(triage.TRIAGE_MESSAGE_KEY))

  def test_similar_testcase_with_issue_recently_closed(self):
    """Tests result is true when there is a similar testcase with issue closed
    recently."""
    self.issue.status = 'Fixed'
    self.issue._monorail_issue.open = False
    self.issue._monorail_issue.closed = (
        test_utils.CURRENT_TIME -
        datetime.timedelta(hours=data_types.MIN_ELAPSED_TIME_SINCE_FIXED - 1))
    self.issue.save()

    similar_testcase = test_utils.create_generic_testcase()
    similar_testcase.one_time_crasher_flag = False
    similar_testcase.open = False
    similar_testcase.bug_information = str(self.issue.id)
    similar_testcase.put()

    self.assertEqual(
        True,
        triage._check_and_update_similar_bug(self.testcase, self.issue_tracker))

    testcase = data_handler.get_testcase_by_id(self.testcase.key.id())
    self.assertEqual(
        'Delaying filing a bug since similar testcase (2) in issue (1) '
        'was just fixed.', testcase.get_metadata(triage.TRIAGE_MESSAGE_KEY))



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Run command based on the current task."""

import functools
import sys
import time

import six

from clusterfuzz._internal.base import errors
from clusterfuzz._internal.base import tasks
from clusterfuzz._internal.base import utils
from clusterfuzz._internal.bot.tasks import analyze_task
from clusterfuzz._internal.bot.tasks import blame_task
from clusterfuzz._internal.bot.tasks import corpus_pruning_task
from clusterfuzz._internal.bot.tasks import fuzz_task
from clusterfuzz._internal.bot.tasks import impact_task
from clusterfuzz._internal.bot.tasks import minimize_task
from clusterfuzz._internal.bot.tasks import progression_task
from clusterfuzz._internal.bot.tasks import regression_task
from clusterfuzz._internal.bot.tasks import symbolize_task
from clusterfuzz._internal.bot.tasks import train_rnn_generator_task
from clusterfuzz._internal.bot.tasks import unpack_task
from clusterfuzz._internal.bot.tasks import upload_reports_task
from clusterfuzz._internal.bot.tasks import variant_task
from clusterfuzz._internal.bot.webserver import http_server
from clusterfuzz._internal.datastore import data_handler
from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.metrics import logs
from clusterfuzz._internal.system import environment
from clusterfuzz._internal.system import process_handler
from clusterfuzz._internal.system import shell

COMMAND_MAP = {
    'analyze': analyze_task,
    'blame': blame_task,
    'corpus_pruning': corpus_pruning_task,
    'fuzz': fuzz_task,
    'impact': impact_task,
    'minimize': minimize_task,
    'train_rnn_generator': train_rnn_generator_task,
    'progression': progression_task,
    'regression': regression_task,
    'symbolize': symbolize_task,
    'unpack': unpack_task,
    'upload_reports': upload_reports_task,
    'variant': variant_task,
}

TASK_RETRY_WAIT_LIMIT = 5 * 60  # 5 minutes.


class Error(Exception):
  """Base commands exceptions."""


class AlreadyRunningError(Error):
  """Exception raised for a task that is already running on another bot."""


def cleanup_task_state():
  """Cleans state before and after a task is executed."""
  # Cleanup stale processes.
  process_handler.cleanup_stale_processes()

  # Clear build urls, temp and testcase directories.
  shell.clear_build_urls_directory()
  shell.clear_crash_stacktraces_directory()
  shell.clear_testcase_directories()
  shell.clear_temp_directory()
  shell.clear_system_temp_directory()
  shell.clear_device_temp_directories()

  # Reset memory tool environment variables.
  environment.reset_current_memory_tool_options()

  # Call python's garbage collector.
  utils.python_gc()


def is_supported_cpu_arch_for_job():
  """Return true if the current cpu architecture can run this job."""
  cpu_arch = environment.get_cpu_arch()
  if not cpu_arch:
    # No cpu architecture check is defined for this platform, bail out.
    return True

  supported_cpu_arch = environment.get_value('CPU_ARCH')
  if not supported_cpu_arch:
    # No specific cpu architecture requirement specified in job, bail out.
    return True

  # Convert to list just in case anyone specifies value as a single string.
  supported_cpu_arch_list = list(supported_cpu_arch)

  return cpu_arch in supported_cpu_arch_list


def update_environment_for_job(environment_string):
  """Process the environment variable string included with a job."""
  # Now parse the job's environment definition.
  environment_values = (
      environment.parse_environment_definition(environment_string))

  for key, value in six.iteritems(environment_values):
    environment.set_value(key, value)

  # If we share the build with another job type, force us to be a custom binary
  # job type.
  if environment.get_value('SHARE_BUILD_WITH_JOB_TYPE'):
    environment.set_value('CUSTOM_BINARY', True)

  # Allow the default FUZZ_TEST_TIMEOUT and MAX_TESTCASES to be overridden on
  # machines that are preempted more often.
  fuzz_test_timeout_override = environment.get_value(
      'FUZZ_TEST_TIMEOUT_OVERRIDE')
  if fuzz_test_timeout_override:
    environment.set_value('FUZZ_TEST_TIMEOUT', fuzz_test_timeout_override)

  max_testcases_override = environment.get_value('MAX_TESTCASES_OVERRIDE')
  if max_testcases_override:
    environment.set_value('MAX_TESTCASES', max_testcases_override)

  if environment.is_trusted_host():
    environment_values['JOB_NAME'] = environment.get_value('JOB_NAME')
    from clusterfuzz._internal.bot.untrusted_runner import \
        environment as worker_environment
    worker_environment.update_environment(environment_values)


def set_task_payload(func):
  """Set TASK_PAYLOAD and unset TASK_PAYLOAD."""

  @functools.wraps(func)
  def wrapper(task):
    """Wrapper."""
    environment.set_value('TASK_PAYLOAD', task.payload())
    try:
      return func(task)
    except:  # Truly catch *all* exceptions.
      e = sys.exc_info()[1]
      e.extras = {'task_payload': environment.get_value('TASK_PAYLOAD')}
      raise
    finally:
      environment.remove_key('TASK_PAYLOAD')

  return wrapper


def should_update_task_status(task_name):
  """Whether the task status should be automatically handled."""
  return task_name not in [
      # Multiple fuzz tasks are expected to run in parallel.
      'fuzz',

      # The task payload can't be used as-is for de-duplication purposes as it
      # includes revision. corpus_pruning_task calls update_task_status itself
      # to handle this.
      # TODO(ochang): This will be cleaned up as part of migration to Pub/Sub.
      'corpus_pruning',
  ]


def start_web_server_if_needed():
  """Start web server for blackbox fuzzer jobs (non-engine fuzzer jobs)."""
  if environment.is_engine_fuzzer_job():
    return

  try:
    http_server.start()
  except Exception:
    logs.log_error('Failed to start web server, skipping.')


def run_command(task_name, task_argument, job_name):
  """Run the command."""
  if task_name not in COMMAND_MAP:
    logs.log_error("Unknown command '%s'" % task_name)
    return

  task_module = COMMAND_MAP[task_name]

  # If applicable, ensure this is the only instance of the task running.
  task_state_name = ' '.join([task_name, task_argument, job_name])
  if should_update_task_status(task_name):
    if not data_handler.update_task_status(task_state_name,
                                           data_types.TaskState.STARTED):
      logs.log('Another instance of "{}" already '
               'running, exiting.'.format(task_state_name))
      raise AlreadyRunningError

  try:
    task_module.execute_task(task_argument, job_name)
  except errors.InvalidTestcaseError:
    # It is difficult to try to handle the case where a test case is deleted
    # during processing. Rather than trying to catch by checking every point
    # where a test case is reloaded from the datastore, just abort the task.
    logs.log_warn('Test case %s no longer exists.' % task_argument)
  except BaseException:
    # On any other exceptions, update state to reflect error and re-raise.
    if should_update_task_status(task_name):
      data_handler.update_task_status(task_state_name,
                                      data_types.TaskState.ERROR)

    raise

  # Task completed successfully.
  if should_update_task_status(task_name):
    data_handler.update_task_status(task_state_name,
                                    data_types.TaskState.FINISHED)


# pylint: disable=too-many-nested-blocks
# TODO(mbarbella): Rewrite this function to avoid nesting issues.
@set_task_payload
def process_command(task):
  """Figures out what to do with the given task and executes the command."""
  logs.log("Executing command '%s'" % task.payload())
  if not task.payload().strip():
    logs.log_error('Empty task received.')
    return

  # Parse task payload.
  task_name = task.command
  task_argument = task.argument
  job_name = task.job

  environment.set_value('TASK_NAME', task_name)
  environment.set_value('TASK_ARGUMENT', task_argument)
  environment.set_value('JOB_NAME', job_name)
  if job_name != 'none':
    job = data_types.Job.query(data_types.Job.name == job_name).get()
    # Job might be removed. In that case, we don't want an exception
    # raised and causing this task to be retried by another bot.
    if not job:
      logs.log_error("Job '%s' not found." % job_name)
      return

    if not job.platform:
      error_string = "No platform set for job '%s'" % job_name
      logs.log_error(error_string)
      raise errors.BadStateError(error_string)

    # A misconfiguration led to this point. Clean up the job if necessary.
    job_queue_suffix = tasks.queue_suffix_for_platform(job.platform)
    bot_queue_suffix = tasks.default_queue_suffix()

    if job_queue_suffix != bot_queue_suffix:
      # This happens rarely, store this as a hard exception.
      logs.log_error(
          'Wrong platform for job %s: job queue [%s], bot queue [%s].' %
          (job_name, job_queue_suffix, bot_queue_suffix))

      # Try to recreate the job in the correct task queue.
      new_queue = (
          tasks.high_end_queue() if task.high_end else tasks.regular_queue())
      new_queue += job_queue_suffix

      # Command override is continuously run by a bot. If we keep failing
      # and recreating the task, it will just DoS the entire task queue.
      # So, we don't create any new tasks in that case since it needs
      # manual intervention to fix the override anyway.
      if not task.is_command_override:
        try:
          tasks.add_task(task_name, task_argument, job_name, new_queue)
        except Exception:
          # This can happen on trying to publish on a non-existent topic, e.g.
          # a topic for a high-end bot on another platform. In this case, just
          # give up.
          logs.log_error('Failed to fix platform and re-add task.')

      # Add a wait interval to avoid overflowing task creation.
      failure_wait_interval = environment.get_value('FAIL_WAIT')
      time.sleep(failure_wait_interval)
      return

    if task_name != 'fuzz':
      # Make sure that our platform id matches that of the testcase (for
      # non-fuzz tasks).
      testcase = data_handler.get_entity_by_type_and_id(data_types.Testcase,
                                                        task_argument)
      if testcase:
        current_platform_id = environment.get_platform_id()
        testcase_platform_id = testcase.platform_id

        # This indicates we are trying to run this job on the wrong platform.
        # This can happen when you have different type of devices (e.g
        # android) on the same platform group. In this case, we just recreate
        # the task.
        if (task_name != 'variant' and testcase_platform_id and
            not utils.fields_match(testcase_platform_id, current_platform_id)):
          logs.log(
              'Testcase %d platform (%s) does not match with ours (%s), exiting'
              % (testcase.key.id(), testcase_platform_id, current_platform_id))
          tasks.add_task(
              task_name,
              task_argument,
              job_name,
              wait_time=utils.random_number(1, TASK_RETRY_WAIT_LIMIT))
          return

    # Some fuzzers contain additional environment variables that should be
    # set for them. Append these for tests generated by these fuzzers and for
    # the fuzz command itself.
    fuzzer_name = None
    if task_name == 'fuzz':
      fuzzer_name = task_argument
    elif testcase:
      fuzzer_name = testcase.fuzzer_name

    # Get job's environment string.
    environment_string = job.get_environment_string()

    if task_name == 'minimize':
      # Let jobs specify a different job and fuzzer to minimize with.
      job_environment = job.get_environment()
      minimize_job_override = job_environment.get('MINIMIZE_JOB_OVERRIDE')
      if minimize_job_override:
        minimize_job = data_types.Job.query(
            data_types.Job.name == minimize_job_override).get()
        if minimize_job:
          environment.set_value('JOB_NAME', minimize_job_override)
          environment_string = minimize_job.get_environment_string()
          environment_string += '\nORIGINAL_JOB_NAME = %s\n' % job_name
          job_name = minimize_job_override
        else:
          logs.log_error(
              'Job for minimization not found: %s.' % minimize_job_override)
          # Fallback to using own job for minimization.

      minimize_fuzzer_override = job_environment.get('MINIMIZE_FUZZER_OVERRIDE')
      fuzzer_name = minimize_fuzzer_override or fuzzer_name

    if fuzzer_name and not environment.is_engine_fuzzer_job(job_name):
      fuzzer = data_types.Fuzzer.query(
          data_types.Fuzzer.name == fuzzer_name).get()
      additional_default_variables = ''
      additional_variables_for_job = ''
      if (fuzzer and hasattr(fuzzer, 'additional_environment_string') and
          fuzzer.additional_environment_string):
        for line in fuzzer.additional_environment_string.splitlines():
          # Job specific values may be defined in fuzzer additional
          # environment variable name strings in the form
          # job_name:VAR_NAME = VALUE.
          if '=' in line and ':' in line.split('=', 1)[0]:
            fuzzer_job_name, environment_definition = line.split(':', 1)
            if fuzzer_job_name == job_name:
              additional_variables_for_job += '\n%s' % environment_definition
            continue

          additional_default_variables += '\n%s' % line

      environment_string += additional_default_variables
      environment_string += additional_variables_for_job

    # Update environment for the job.
    update_environment_for_job(environment_string)

  # Match the cpu architecture with the ones required in the job definition.
  # If they don't match, then bail out and recreate task.
  if not is_supported_cpu_arch_for_job():
    logs.log(
        'Unsupported cpu architecture specified in job definition, exiting.')
    tasks.add_task(
        task_name,
        task_argument,
        job_name,
        wait_time=utils.random_number(1, TASK_RETRY_WAIT_LIMIT))
    return

  # Initial cleanup.
  cleanup_task_state()

  start_web_server_if_needed()

  try:
    run_command(task_name, task_argument, job_name)
  finally:
    # Final clean up.
    cleanup_task_state()



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""File operations host (client)."""

import os
import shutil

from clusterfuzz._internal.metrics import logs
from clusterfuzz._internal.protos import untrusted_runner_pb2
from clusterfuzz._internal.system import environment
from clusterfuzz._internal.system import shell

from . import file_utils
from . import host


def is_directory_parent(path, directory):
  """Check whether if |directory| is a parent of |path|."""
  path = os.path.abspath(path)
  directory = os.path.abspath(directory)

  path_components = path.split(os.sep)
  directory_components = directory.split(os.sep)

  if len(path_components) <= len(directory_components):
    return False

  return all(path_components[i] == directory_components[i]
             for i in range(len(directory_components)))


def _rebase(path, target_base, cur_base):
  """Rebase a path."""
  if not path:
    # Don't rebase if the path is None or empty string (in case of default
    # variable value).
    return path

  if os.path.abspath(path).startswith(target_base):
    # Already rebased.
    return path

  rel_path = os.path.relpath(os.path.abspath(path), cur_base)

  if rel_path == os.curdir:
    return target_base

  # Only paths relative to ROOT_DIR are supported.
  assert not rel_path.startswith(os.pardir), 'Bad relative path %s' % rel_path
  return os.path.join(target_base, rel_path)


def rebase_to_host_root(worker_path):
  """Return corresponding host root given a worker CF path."""
  return _rebase(worker_path, environment.get_value('ROOT_DIR'),
                 environment.get_value('WORKER_ROOT_DIR'))


def rebase_to_worker_root(host_path):
  """Return corresponding worker path given a host CF path."""
  return _rebase(host_path, environment.get_value('WORKER_ROOT_DIR'),
                 environment.get_value('ROOT_DIR'))


def create_directory(path, create_intermediates=False):
  """Create a directory."""
  request = untrusted_runner_pb2.CreateDirectoryRequest(
      path=path, create_intermediates=create_intermediates)

  response = host.stub().CreateDirectory(request)
  return response.result


def remove_directory(path, recreate=False):
  """Remove a directory. If |recreate| is set, always creates the directory even
  if it did not exist."""
  request = untrusted_runner_pb2.RemoveDirectoryRequest(
      path=path, recreate=recreate)

  response = host.stub().RemoveDirectory(request)
  return response.result


def list_files(path, recursive=False):
  """List files in the directory. Returns full file paths."""
  request = untrusted_runner_pb2.ListFilesRequest(
      path=path, recursive=recursive)

  response = host.stub().ListFiles(request)
  return response.file_paths


def copy_file_to_worker(host_path, worker_path):
  """Copy file from host to worker. |worker_path| must be a full path (including
  the filename). Any directories will be created if needed."""
  with open(host_path, 'rb') as f:
    request_iterator = file_utils.file_chunk_generator(f)
    metadata = [('path-bin', worker_path.encode('utf-8'))]

    response = host.stub().CopyFileTo(request_iterator, metadata=metadata)
    return response.result


def write_data_to_worker(data, worker_path):
  """Write data to a file on the worker."""
  request_iterator = file_utils.data_chunk_generator(data)
  metadata = [('path-bin', worker_path.encode('utf-8'))]

  response = host.stub().CopyFileTo(request_iterator, metadata=metadata)
  return response.result


def copy_file_from_worker(worker_path, host_path):
  """Copy file from worker to host."""
  request = untrusted_runner_pb2.CopyFileFromRequest(path=worker_path)
  response = host.stub().CopyFileFrom(request)
  file_utils.write_chunks(host_path, response)
  metadata = dict(response.trailing_metadata())
  if metadata.get('result') != 'ok':
    # file_utils.write_chunks always opens the file for writing, so remove it
    # here.
    os.remove(host_path)
    return False

  return True


def copy_directory_to_worker(host_directory, worker_directory, replace=False):
  """Recursively copy a directory to the worker. Directories are created as
  needed. Unless |replace| is True, files already in |worker_path| will remain
  after this call."""
  if replace:
    remove_directory(worker_directory, recreate=True)

  for root, _, files in shell.walk(host_directory):
    for filename in files:
      file_path = os.path.join(root, filename)
      worker_file_path = os.path.join(
          worker_directory, os.path.relpath(file_path, host_directory))
      if not copy_file_to_worker(file_path, worker_file_path):
        logs.log_warn('Failed to copy %s to worker.' % file_path)
        return False

  return True


def copy_directory_from_worker(worker_directory, host_directory, replace=False):
  """Recursively copy a directory from the worker. Directories are created as
  needed. Unless |replace| is True, files already in |host_directory| will
  remain after this call."""
  if replace and os.path.exists(host_directory):
    shutil.rmtree(host_directory, ignore_errors=True)
    os.mkdir(host_directory)

  for worker_file_path in list_files(worker_directory, recursive=True):
    relative_worker_file_path = os.path.relpath(worker_file_path,
                                                worker_directory)
    host_file_path = os.path.join(host_directory, relative_worker_file_path)

    # Be careful with the path provided by the worker here. We want to make sure
    # we're only writing files to |host_directory| and not outside it.
    if not is_directory_parent(host_file_path, host_directory):
      logs.log_warn('copy_directory_from_worker: Attempt to escape |host_dir|.')
      return False

    host_file_directory = os.path.dirname(host_file_path)
    if not os.path.exists(host_file_directory):
      os.makedirs(host_file_directory)

    if not copy_file_from_worker(worker_file_path, host_file_path):
      logs.log_warn('Failed to copy %s from worker.' % worker_file_path)
      return False

  return True


def stat(path):
  """stat() a path."""
  request = untrusted_runner_pb2.StatRequest(path=path)
  response = host.stub().Stat(request)
  if not response.result:
    return None

  return response


def clear_testcase_directories():
  """Clear the testcases directories on the worker."""
  remove_directory(
      rebase_to_worker_root(environment.get_value('FUZZ_INPUTS')),
      recreate=True)
  remove_directory(
      rebase_to_worker_root(environment.get_value('FUZZ_INPUTS_DISK')),
      recreate=True)


def clear_build_urls_directory():
  """Clear the build urls directory on the worker."""
  remove_directory(
      rebase_to_worker_root(environment.get_value('BUILD_URLS_DIR')),
      recreate=True)


def clear_temp_directory():
  """Clear the temp directory on the worker."""
  remove_directory(environment.get_value('WORKER_BOT_TMPDIR'), recreate=True)


def push_testcases_to_worker():
  """Push all testcases to the worker."""
  local_testcases_directory = environment.get_value('FUZZ_INPUTS')
  worker_testcases_directory = rebase_to_worker_root(local_testcases_directory)
  return copy_directory_to_worker(
      local_testcases_directory, worker_testcases_directory, replace=True)


def pull_testcases_from_worker():
  """Pull all testcases to the worker."""
  local_testcases_directory = environment.get_value('FUZZ_INPUTS')
  worker_testcases_directory = rebase_to_worker_root(local_testcases_directory)
  return copy_directory_from_worker(
      worker_testcases_directory, local_testcases_directory, replace=True)


def get_fuzz_targets(path):
  """Get list of fuzz target paths."""
  request = untrusted_runner_pb2.GetFuzzTargetsRequest(path=path)
  response = host.stub().GetFuzzTargets(request)
  return response.fuzz_target_paths



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""File operations implementations."""

import os

from clusterfuzz._internal.bot.fuzzers import utils as fuzzers_utils
from clusterfuzz._internal.protos import untrusted_runner_pb2
from clusterfuzz._internal.system import shell

from . import file_utils


def create_directory(request, _):
  """Create a directory."""
  result = shell.create_directory(request.path, request.create_intermediates)
  return untrusted_runner_pb2.CreateDirectoryResponse(result=result)


def remove_directory(request, _):
  """Remove a directory."""
  result = shell.remove_directory(request.path, request.recreate)
  return untrusted_runner_pb2.RemoveDirectoryResponse(result=result)


def list_files(request, _):
  """List files."""
  file_paths = []
  if request.recursive:
    for root, _, files in shell.walk(request.path):
      for filename in files:
        file_paths.append(os.path.join(root, filename))
  else:
    file_paths.extend(
        os.path.join(request.path, path) for path in os.listdir(request.path))

  return untrusted_runner_pb2.ListFilesResponse(file_paths=file_paths)


def copy_file_to_worker(request_iterator, context):
  """Copy file from host to worker."""
  metadata = dict(context.invocation_metadata())
  path = metadata['path-bin'].decode('utf-8')

  # Create intermediate directories if needed.
  directory = os.path.dirname(path)
  if not os.path.exists(directory):
    try:
      os.makedirs(directory)
    except Exception:
      pass

  if not os.path.isdir(directory):
    # Failed to create intermediate directories.
    return untrusted_runner_pb2.CopyFileToResponse(result=False)

  file_utils.write_chunks(path, request_iterator)
  return untrusted_runner_pb2.CopyFileToResponse(result=True)


def copy_file_from_worker(request, context):
  """Copy file from worker to host."""
  path = request.path
  if not os.path.isfile(path):
    context.set_trailing_metadata([('result', 'invalid-path')])
    return

  with open(path, 'rb') as f:
    for chunk in file_utils.file_chunk_generator(f):
      yield chunk
  context.set_trailing_metadata([('result', 'ok')])


def stat(request, _):
  """Stat a path."""
  if not os.path.exists(request.path):
    return untrusted_runner_pb2.StatResponse(result=False)

  stat_result = os.stat(request.path)
  return untrusted_runner_pb2.StatResponse(
      result=True,
      st_mode=stat_result.st_mode,
      st_size=stat_result.st_size,
      st_atime=stat_result.st_atime,
      st_mtime=stat_result.st_mtime,
      st_ctime=stat_result.st_ctime)


def get_fuzz_targets(request, _):
  """Get list of fuzz targets."""
  fuzz_target_paths = fuzzers_utils.get_fuzz_targets_local(request.path)
  return untrusted_runner_pb2.GetFuzzTargetsResponse(
      fuzz_target_paths=fuzz_target_paths)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Wifi related functions."""

import os
import time

try:
  from shlex import quote
except ImportError:
  from pipes import quote

from clusterfuzz._internal.config import db_config
from clusterfuzz._internal.metrics import logs
from clusterfuzz._internal.system import environment

from . import adb
from . import app

WIFI_UTIL_PACKAGE_NAME = 'com.android.tradefed.utils.wifi'
WIFI_UTIL_CALL_PATH = '%s/.WifiUtil' % WIFI_UTIL_PACKAGE_NAME


def disable():
  """Disable wifi."""
  adb.run_shell_command(['svc', 'wifi', 'disable'])


def enable():
  """Enable wifi."""
  adb.run_shell_command(['svc', 'wifi', 'enable'])


def disable_airplane_mode():
  """Disable airplane mode."""
  adb.run_shell_command(['settings', 'put', 'global', 'airplane_mode_on', '0'])
  adb.run_shell_command([
      'am', 'broadcast', '-a', 'android.intent.action.AIRPLANE_MODE', '--ez',
      'state', 'false'
  ])


def configure(force_enable=False):
  """Configure airplane mode and wifi on device."""
  # The reproduce tool shouldn't inherit wifi settings from jobs.
  if environment.get_value('REPRODUCE_TOOL'):
    return

  # Airplane mode should be disabled in all cases. This can get inadvertently
  # turned on via gestures.
  disable_airplane_mode()

  # Need to disable wifi before changing configuration.
  disable()

  # Check if wifi needs to be enabled. If not, then no need to modify the
  # supplicant file.
  wifi_enabled = force_enable or environment.get_value('WIFI', True)
  if not wifi_enabled:
    # No more work to do, we already disabled it at start.
    return

  # Wait 2 seconds to allow the wifi to be enabled.
  enable()
  time.sleep(2)

  # Install helper apk to configure wifi.
  wifi_util_apk_path = os.path.join(
      environment.get_platform_resources_directory(), 'wifi_util.apk')
  if not app.is_installed(WIFI_UTIL_PACKAGE_NAME):
    app.install(wifi_util_apk_path)

  # Get ssid and password from admin configuration.
  if environment.is_android_cuttlefish():
    wifi_ssid = 'VirtWifi'
    wifi_password = ''
  else:
    config = db_config.get()
    if not config.wifi_ssid:
      logs.log('No wifi ssid is set, skipping wifi config.')
      return
    wifi_ssid = config.wifi_ssid
    wifi_password = config.wifi_password or ''

  connect_wifi_command = (
      'am instrument -e method connectToNetwork -e ssid {ssid} ')
  if wifi_password:
    connect_wifi_command += '-e psk {password} '
  connect_wifi_command += '-w {call_path}'

  output = adb.run_shell_command(
      connect_wifi_command.format(
          ssid=quote(wifi_ssid),
          password=quote(wifi_password),
          call_path=WIFI_UTIL_CALL_PATH))
  if 'result=true' not in output:
    logs.log_warn('Failed to connect to wifi.', output=output)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""js_unittest.py runs JS tests under src/appengine"""

import os
import re
import sys
import time

from selenium import webdriver

from local.butler import common

_SUITE_SEPARATOR = '=' * 80
_TEST_SEPARATOR = '-' * 80


def _parse_error_report(driver):
  """Parse failed test report from Mocha HTML result"""
  error_report = ''

  # Remove the replay buttons next to test names
  for elem in driver.find_elements_by_css_selector('#mocha-report .suite h2 a'):
    driver.execute_script('arguments[0].remove()', elem)

  suites = driver.find_elements_by_css_selector('#mocha-report .suite .suite')
  for suite in suites:
    failed_tests = suite.find_elements_by_css_selector('.test.fail')
    if not failed_tests:
      continue

    suite_name = suite.find_element_by_css_selector('h1').text.strip()

    error_report += '\n\n%s\n' % _SUITE_SEPARATOR
    error_report += '%s\n' % suite_name

    for failed_test in failed_tests:
      name = failed_test.find_element_by_css_selector('h2').text.strip()
      trace = failed_test.find_element_by_css_selector('.error').text.strip()
      trace = re.sub('^', '| ', trace)
      trace = re.sub('\n', '\n| ', trace)

      error_report += '%s\n' % _TEST_SEPARATOR
      error_report += 'Failed test: %s\n' % name
      error_report += '%s\n' % trace

  return error_report


def execute(args):
  """Run Javascript unit tests. Here are the steps:

     1. Execute the HTML with chromedriver.
     2. Read the test result from the HTML."""
  test_filepath = os.path.join('src', 'appengine', 'private', 'test.html')
  print('Running chromedriver on %s' % test_filepath)

  chrome_options = webdriver.ChromeOptions()
  chrome_options.add_argument('--allow-file-access-from-files')

  is_ci = os.getenv('TEST_BOT_ENVIRONMENT')
  if is_ci:
    # Turn off sandbox since running under root, with trusted tests.
    chrome_options.add_argument('--no-sandbox')
    chrome_options.add_argument('--headless')

  driver = webdriver.Chrome(
      executable_path=common.get_chromedriver_path(),
      chrome_options=chrome_options)

  try:
    driver.get('file://%s' % os.path.abspath(test_filepath))

    # Wait for tests to be completed.
    while True:
      success_count = driver.execute_script(
          'return WCT._reporter.stats.passes;')
      failure_count = driver.execute_script(
          'return WCT._reporter.stats.failures;')
      sys.stdout.write(
          '\rSuccess: %d, Failure: %d' % (success_count, failure_count))
      sys.stdout.flush()

      is_complete = driver.execute_script('return WCT._reporter.complete;')
      if is_complete:
        break

      time.sleep(0.1)

    sys.stdout.write('\r' + (' ' * 70))
    sys.stdout.flush()

    success_count = int(
        driver.find_element_by_css_selector('#mocha-stats .passes em').text)
    failure_count = int(
        driver.find_element_by_css_selector('#mocha-stats .failures em').text)
    error_report = _parse_error_report(driver)

    if error_report:
      print(error_report)

    print()
    print(_SUITE_SEPARATOR)
    print('Test results:')
    print('| Success: %d' % success_count)
    print('| Failure: %d' % failure_count)
    print(_SUITE_SEPARATOR)
    print()

    if args.persist:
      # pylint: disable=eval-used
      eval(
          input('--persist is used. Leave the browser open.'
                ' Press ENTER to close it:'))
  finally:
    driver.quit()

  if failure_count > 0:
    sys.exit(1)



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Helper functions for running commands on emulated Android devices."""

import os
import re
import subprocess

from clusterfuzz._internal.metrics import logs
from clusterfuzz._internal.platforms.android import adb
from clusterfuzz._internal.system import environment

try:
  from clusterfuzz._internal.system import new_process
except ImportError:
  # On App Engine.
  new_process = None

# Output pattern to parse stdout for serial number
DEVICE_SERIAL_RE = re.compile(r'DEVICE_SERIAL: (.+)')


class EmulatorError(Exception):
  """Error for errors handling the Android emulator."""


class EmulatorProcess(object):
  """A EmulatorProcess encapsulates the creation, running, and destruction
  of Android emulator processes."""

  def __init__(self):
    self.process_runner = None
    self.process = None

  def create(self, work_dir):
    """Configures a emulator process which can subsequently be `run`."""
    self.process_runner = new_process.ProcessRunner(
        os.path.join(work_dir, '../emulator/run'))

  def run(self):
    """Actually runs a emulator, assuming `create` has already been called."""
    if not self.process_runner:
      raise EmulatorError('Attempted to `run` emulator before calling `create`')

    logs.log('Starting emulator.')
    self.process = self.process_runner.run(
        stdout=subprocess.PIPE, stderr=subprocess.DEVNULL)

    device_serial = None
    while not device_serial:
      line = self.process.popen.stdout.readline().decode()
      match = DEVICE_SERIAL_RE.match(line)
      if match:
        device_serial = match.group(1)

    # Close the pipe so we don't hang.
    self.process.popen.stdout.close()

    logs.log('Found serial ID: %s.' % device_serial)
    environment.set_value('ANDROID_SERIAL', device_serial)

    logs.log('Waiting on device')
    adb.wait_for_device()
    logs.log('Device is online')

  def kill(self):
    """ Kills the currently-running emulator, if there is one. """
    if self.process:
      logs.log('Stopping emulator.')
      self.process.kill()
      self.process = None



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Issue tracker policy."""

from collections import namedtuple

from clusterfuzz._internal.config import local_config

Status = namedtuple('Status',
                    ['assigned', 'duplicate', 'wontfix', 'fixed', 'verified'])

EXPECTED_STATUSES = [
    'assigned',
    'duplicate',
    'wontfix',
    'fixed',
    'verified',
    'new',
]


class ConfigurationError(Exception):
  """Base configuration error class."""


class NewIssuePolicy(object):
  """New issue policy."""

  def __init__(self):
    self.status = ''
    self.ccs = []
    self.labels = []
    self.issue_body_footer = ''


def _to_str_list(values):
  """Convert a list to a list of strs."""
  return [str(value) for value in values]


class IssueTrackerPolicy(object):
  """Represents an issue tracker policy."""

  def __init__(self, data):
    self._data = data
    if 'status' not in self._data:
      raise ConfigurationError('Status not set in policies.')

    if 'labels' not in self._data:
      raise ConfigurationError('Labels not set in policies.')

    for status in EXPECTED_STATUSES:
      if status not in self._data['status']:
        raise ConfigurationError(
            'Expected status {} is not set.'.format(status))

  def status(self, status_type):
    """Get the actual status string for the given type."""
    return self._data['status'][status_type]

  def label(self, label_type):
    """Get the actual label string for the given type."""
    label = self._data['labels'].get(label_type)
    if label is None:
      return None

    return str(label)

  def substitution_mapping(self, label):
    """Get an explicit substitution mapping."""
    if 'substitutions' not in self._data:
      return label

    mapped = self._data['substitutions'].get(label)
    if not mapped:
      return label

    return str(mapped)

  @property
  def deadline_policy_message(self):
    """Get the deadline policy message, if it exists."""
    return self._data.get('deadline_policy_message')

  @property
  def unreproducible_component(self):
    """Get the component for unreproducible bugs, if it exists."""
    return self._data.get('unreproducible_component')

  def get_new_issue_properties(self, is_security, is_crash):
    """Get the properties to apply to a new issue."""
    policy = NewIssuePolicy()

    if 'all' in self._data:
      self._apply_new_issue_properties(policy, self._data['all'], is_crash)

    if is_security:
      if 'security' in self._data:
        self._apply_new_issue_properties(policy, self._data['security'],
                                         is_crash)
    else:
      if 'non_security' in self._data:
        self._apply_new_issue_properties(policy, self._data['non_security'],
                                         is_crash)

    return policy

  def _apply_new_issue_properties(self, policy, issue_type, is_crash):
    """Apply issue policies."""
    if not issue_type:
      return

    if 'status' in issue_type:
      policy.status = self._data['status'][issue_type['status']]

    if 'ccs' in issue_type:
      policy.labels.extend(issue_type['ccs'])

    labels = issue_type.get('labels')
    if labels:
      policy.labels.extend(_to_str_list(labels))

    issue_body_footer = issue_type.get('issue_body_footer')
    if issue_body_footer:
      policy.issue_body_footer = issue_body_footer

    if is_crash:
      crash_labels = issue_type.get('crash_labels')
      if crash_labels:
        policy.labels.extend(_to_str_list(crash_labels))
    else:
      non_crash_labels = issue_type.get('non_crash_labels')
      if non_crash_labels:
        policy.labels.extend(_to_str_list(non_crash_labels))

  def get_existing_issue_properties(self):
    """Get the properties to apply to a new issue."""
    policy = NewIssuePolicy()

    if 'existing' in self._data:
      self._apply_new_issue_properties(policy, self._data['existing'], False)

    return policy


def get(project_name):
  """Get policy."""
  issue_tracker_config = local_config.IssueTrackerConfig()
  project_config = issue_tracker_config.get(project_name)
  if not project_config:
    raise ConfigurationError(
        'Issue tracker for {} does not exist'.format(project_name))

  if not 'policies' in project_config:
    raise ConfigurationError(
        'Policies for {} do not exist'.format(project_name))

  return IssueTrackerPolicy(project_config['policies'])


def get_empty():
  """Get an empty policy."""
  return IssueTrackerPolicy({
      'status': {
          'assigned': 'unused',
          'duplicate': 'unused',
          'wontfix': 'unused',
          'fixed': 'unused',
          'verified': 'unused',
          'new': 'unused',
      },
      'labels': {},
  })



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for the Mac initialization script."""

import unittest

import mock

from clusterfuzz._internal.bot.init_scripts import mac
from clusterfuzz._internal.tests.test_libs import helpers


class RunTest(unittest.TestCase):
  """Test run."""

  def setUp(self):
    helpers.patch(self, [
        'clusterfuzz._internal.bot.init_scripts.init_runner.run',
        'os.path.expanduser',
        'clusterfuzz._internal.system.shell.remove_directory',
        'shutil.rmtree',
        'subprocess.Popen',
        'os.path.exists',
    ])
    self.popen = mock.Mock()
    self.stdout = []

    def readline():
      return self.stdout.pop(0)

    self.popen.stdout.readline = readline
    self.mock.Popen.return_value = self.popen

    def expanduser(path):
      return path.replace('~', '/Users/chrome-bot')

    self.mock.expanduser.side_effect = expanduser

  def test_run(self):
    """Test run."""
    self.stdout = [
        b'aaaa\n', b'bbbb\n',
        (b'Path: /var/folders/bg/tn9j_qb532s4fz11rzz7m6sc0000gm/0'
         b'//com.apple.LaunchServices-134500.csstore\n'), b'cccc\n', b''
    ]
    mac.run()

    self.mock.exists.return_value = True
    self.mock.rmtree.assert_has_calls([
        mock.call(
            '/var/folders/bg/tn9j_qb532s4fz11rzz7m6sc0000gm/0',
            ignore_errors=True),
        mock.call(
            '/var/folders/bg/tn9j_qb532s4fz11rzz7m6sc0000gm/T',
            ignore_errors=True)
    ])



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for the stack frame module."""

import re
import unittest

from clusterfuzz._internal.crash_analysis.stack_parsing import stack_parser


class StackFrameTestcase(unittest.TestCase):
  """Stack analyzer tests."""

  def test_parse_stack_frame(self):
    """Check that stackframes are correctly parsed."""
    spec = stack_parser.StackFrameSpec(
        address=1,
        filename=[4, 8],
        fileline=[5, 9],
    )
    stack_frame_regex = re.compile(
        r'([xXa-fA-F0-9]*) '
        # optional parts: filename:line:char or filename:line
        r'(((\w+):(\d+):(\d+))|((\w+):(\d+)))')
    stack_frame_lines = [
        '0xa1c0 a_file:1234:56',
        '0xa1c0 a_file:1234',
    ]
    stack_frames = []
    for stack_frame_line in stack_frame_lines:
      m = stack_frame_regex.match(stack_frame_line)
      stack_frames.append(spec.parse_stack_frame(m))

    for stack_frame in stack_frames:
      self.assertEqual(stack_frame.address, 41408)
      self.assertEqual(stack_frame.filename, 'a_file')
      self.assertEqual(stack_frame.fileline, '1234')

  def test_to_proto(self):
    """Test converting to protobuf."""
    stack_frame = stack_parser.StackFrame(
        address='0x1',
        function_name='fun',
        function_base='0x10',
        function_offset='0x03',
        filename='a_file.cc',
        fileline=42,
        module_name='modname',
        module_base='0x2',
        module_offset='0xbeef',
    )
    stack_frame_proto = stack_frame.to_proto()

    self.assertEqual(stack_frame_proto.instruction, 1)
    self.assertEqual(stack_frame_proto.module.base_address, 2)
    self.assertEqual(stack_frame_proto.module.code_file, 'modname')
    self.assertEqual(stack_frame_proto.function_name, 'fun')
    self.assertEqual(stack_frame_proto.function_base, 16)
    self.assertEqual(stack_frame_proto.source_file_name, 'a_file.cc')
    self.assertEqual(stack_frame_proto.source_line, 42)

  def test_to_proto_big_addresses(self):
    stack_frame = stack_parser.StackFrame(
        address='0xfffffffffffffff2',
        function_base='0xfffffffffffffff3',
        module_base='0x8000000000000000')
    stack_frame_proto = stack_frame.to_proto()

    self.assertEqual(stack_frame_proto.instruction, -14)
    self.assertEqual(stack_frame_proto.function_base, -13)
    self.assertEqual(stack_frame_proto.module.base_address,
                     -9223372036854775808)



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Cron to generate certs for OSS-Fuzz workers."""

from google.cloud import ndb

from clusterfuzz._internal.base import untrusted
from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.metrics import logs
from handlers import base_handler
from libs import handler


def generate_cert(project_name):
  """Generate a self signed cerficate."""
  # Defer imports to avoid issues on Python 2.
  from OpenSSL import crypto

  key = crypto.PKey()
  key.generate_key(crypto.TYPE_RSA, 2048)

  cert = crypto.X509()
  cert.get_subject().C = 'US'
  cert.get_subject().CN = '*' + untrusted.internal_network_domain()
  cert.get_subject().O = project_name
  cert.set_serial_number(9001)
  cert.set_notBefore(b'20000101000000Z')
  cert.set_notAfter(b'21000101000000Z')
  cert.set_issuer(cert.get_subject())
  cert.set_pubkey(key)
  cert.sign(key, 'sha256')

  cert_contents = crypto.dump_certificate(crypto.FILETYPE_PEM, cert)
  key_contents = crypto.dump_privatekey(crypto.FILETYPE_PEM, key)
  return cert_contents, key_contents


class Handler(base_handler.Handler):
  """Generate OSS-Fuzz certs."""

  @handler.cron()
  def get(self):
    """Handles a get request."""
    for project in data_types.OssFuzzProject.query():
      tls_cert_key = ndb.Key(data_types.WorkerTlsCert, project.name)
      if tls_cert_key.get():
        # Already generated.
        continue

      logs.log('Generating cert for %s.' % project.name)
      cert_contents, key_contents = generate_cert(project.name)

      tls_cert = data_types.WorkerTlsCert(
          id=project.name,
          cert_contents=cert_contents,
          key_contents=key_contents)
      tls_cert.put()



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""fuzzer_logs test."""
import datetime
import unittest

import mock

from clusterfuzz._internal.metrics import fuzzer_logs
from clusterfuzz._internal.system import environment
from clusterfuzz._internal.tests.test_libs import helpers as test_helpers


class FuzzerLogsTest(unittest.TestCase):
  """Tests for logs uploading."""

  def setUp(self):
    test_helpers.patch_environ(self)
    environment.set_value('FUZZER_NAME', 'fuzzer_1')
    environment.set_value('JOB_NAME', 'fake_job')

    # To be used for generation of date and time when uploading a log.
    self.fake_utcnow = datetime.datetime(2017, 3, 21, 11, 15, 13, 666666)
    self.fake_log_time = datetime.datetime(2017, 4, 22, 12, 16, 14, 777777)

    test_helpers.patch(self, [
        'datetime.datetime',
        'clusterfuzz._internal.google_cloud_utils.storage.write_data',
    ])

    self.mock.datetime.utcnow.return_value = self.fake_utcnow

  def test_upload_to_logs(self):
    """Test a simple call to upload_to_logs."""
    mock_gsutil = mock.MagicMock()
    self.mock.write_data.return_value = mock_gsutil
    fuzzer_logs.upload_to_logs('fake-gcs-bucket', 'fake content')
    self.mock.write_data.assert_called_once_with(
        'fake content',
        'gs://fake-gcs-bucket/fuzzer_1/fake_job/2017-03-21/11:15:13:666666.log')

  def test_upload_to_logs_with_all_arguments(self):
    """Test a call to upload_to_logs with all arguments being passed."""
    mock_gsutil = mock.MagicMock()
    self.mock.write_data.return_value = mock_gsutil
    fuzzer_logs.upload_to_logs(
        'gcs-bucket',
        'fake content',
        time=self.fake_log_time,
        fuzzer_name='fuzzer_2',
        job_type='another_job')
    self.mock.write_data.assert_called_once_with(
        'fake content',
        'gs://gcs-bucket/fuzzer_2/another_job/2017-04-22/12:16:14:777777.log')



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""update_issue tests."""
import unittest

import flask
import mock
import webtest

from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.tests.test_libs import helpers as test_helpers
from clusterfuzz._internal.tests.test_libs import test_utils
from handlers.testcase_detail import update_issue
from libs import access
from libs import form
from libs.issue_management import issue_tracker_policy
from libs.issue_management import monorail
from libs.issue_management.monorail import issue
from libs.issue_management.monorail import issue_tracker_manager

CHROMIUM_POLICY = issue_tracker_policy.IssueTrackerPolicy({
    'status': {
        'assigned': 'Assigned',
        'duplicate': 'Duplicate',
        'fixed': 'Fixed',
        'new': 'Untriaged',
        'verified': 'Verified',
        'wontfix': 'WontFix'
    },
    'labels': {},
    'existing': {
        'labels': ['Stability-%SANITIZER%']
    },
})


@test_utils.with_cloud_emulators('datastore')
class HandlerTest(unittest.TestCase):
  """Test Handler."""

  def setUp(self):
    test_helpers.patch(self, [
        'clusterfuzz._internal.datastore.data_handler.get_issue_description',
        'clusterfuzz._internal.datastore.data_handler.get_issue_summary',
        'clusterfuzz._internal.datastore.data_handler.get_stacktrace',
        'clusterfuzz._internal.datastore.data_handler.update_group_bug',
        'libs.helpers.get_issue_tracker_for_testcase',
        'libs.auth.get_current_user',
        'handlers.testcase_detail.show.get_testcase_detail',
        'libs.access.get_access',
        'libs.issue_management.issue_tracker_policy.get',
        'libs.issue_management.issue_filer.get_memory_tool_labels',
    ])
    self.mock.get_access.return_value = access.UserAccess.Allowed
    self.mock.get_testcase_detail.return_value = {'testcase': 'yes'}
    self.mock.get_current_user().email = 'test@test.com'
    self.mock.get.return_value = CHROMIUM_POLICY

    flaskapp = flask.Flask('testflask')
    flaskapp.add_url_rule('/', view_func=update_issue.Handler.as_view('/'))
    self.app = webtest.TestApp(flaskapp)

    self.testcase = data_types.Testcase()
    self.testcase.bug_information = ''
    self.testcase.crash_state = 'fake_crash_state'
    self.testcase.put()

  def test_issue_id_not_a_number(self):
    """issue_id is not a number."""
    resp = self.app.post_json(
        '/', {
            'testcaseId': self.testcase.key.id(),
            'issueId': 'aaa',
            'needsSummaryUpdate': '',
            'csrf_token': form.generate_csrf_token(),
        },
        expect_errors=True)

    self.assertEqual(400, resp.status_int)
    self.assertEqual('Issue ID (aaa) is not a number!', resp.json['message'])
    self.assertEqual('test@test.com', resp.json['email'])

  def test_issue_not_found(self):
    """Issue is not found."""
    itm = mock.Mock(spec_set=issue_tracker_manager.IssueTrackerManager)

    self.mock.get_issue_tracker_for_testcase.return_value = (
        monorail.IssueTracker(itm))
    itm.get_issue.return_value = None

    resp = self.app.post_json(
        '/', {
            'testcaseId': self.testcase.key.id(),
            'issueId': '2',
            'needsSummaryUpdate': '',
            'csrf_token': form.generate_csrf_token(),
        },
        expect_errors=True)

    self.assertEqual(404, resp.status_int)
    self.assertEqual('Issue (id=2) is not found!', resp.json['message'])
    self.assertEqual('test@test.com', resp.json['email'])

  def test_issue_not_open(self):
    """Issue is not open."""
    itm = mock.Mock(spec_set=issue_tracker_manager.IssueTrackerManager)
    bug = issue.Issue()
    bug.open = False

    self.mock.get_issue_tracker_for_testcase.return_value = (
        monorail.IssueTracker(itm))
    itm.get_issue.return_value = bug

    resp = self.app.post_json(
        '/', {
            'testcaseId': self.testcase.key.id(),
            'issueId': '2',
            'needsSummaryUpdate': '',
            'csrf_token': form.generate_csrf_token(),
        },
        expect_errors=True)

    self.assertEqual(400, resp.status_int)
    self.assertIn('file a new issue', resp.json['message'])
    self.assertEqual('test@test.com', resp.json['email'])

  def test_succeed(self):
    """Update an issue."""
    bug = issue.Issue()
    bug.open = True
    itm = mock.Mock(project_name='chromium')
    itm.get_issue.return_value = bug

    self.mock.get_issue_tracker_for_testcase.return_value = (
        monorail.IssueTracker(itm))
    self.mock.get_issue_description.return_value = 'description'
    self.mock.get_issue_summary.return_value = 'summary'
    self.mock.get_stacktrace.return_value = 'stacktrace'
    self.mock.get_memory_tool_labels.return_value = ['tool']

    resp = self.app.post_json(
        '/', {
            'testcaseId': self.testcase.key.id(),
            'issueId': '2',
            'needsSummaryUpdate': 'true',
            'csrf_token': form.generate_csrf_token(),
        },
        expect_errors=True)

    self.assertEqual(200, resp.status_int)
    self.assertEqual('yes', resp.json['testcase'])

    self.assertEqual('description', bug.comment)
    self.assertEqual('summary', bug.summary)
    self.assertListEqual(['Stability-tool'], bug.labels)
    self.assertEqual('2', self.testcase.key.get().bug_information)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Remote process host (client)."""

import os
import subprocess

from clusterfuzz._internal.protos import untrusted_runner_pb2
from clusterfuzz._internal.system import new_process
from clusterfuzz._internal.system import process_handler

from . import environment
from . import host


def process_result_from_proto(process_result_proto):
  """Convert ProcessResult proto to new_process.ProcessResult."""
  return new_process.ProcessResult(
      process_result_proto.command, process_result_proto.return_code,
      process_result_proto.output, process_result_proto.time_executed,
      process_result_proto.timed_out)


def run_process(cmdline,
                current_working_directory=None,
                timeout=process_handler.DEFAULT_TEST_TIMEOUT,
                need_shell=False,
                gestures=None,
                env_copy=None,
                testcase_run=True,
                ignore_children=True):
  """Remote version of process_handler.run_process."""
  request = untrusted_runner_pb2.RunProcessRequest(
      cmdline=cmdline,
      current_working_directory=current_working_directory,
      timeout=timeout,
      need_shell=need_shell,
      testcase_run=testcase_run,
      ignore_children=ignore_children)

  if gestures:
    request.gestures.extend(gestures)

  env = {}
  # run_process's local behaviour is to apply the passed |env_copy| on top of
  # the current environment instead of replacing it completely (like with
  # subprocess).
  environment.set_environment_vars(env, os.environ)
  environment.set_environment_vars(env, env_copy)
  request.env_copy.update(env)

  response = host.stub().RunProcess(request)
  return response.return_code, response.execution_time, response.output


class RemoteProcessRunner(new_process.ProcessRunner):
  """Remote child process."""

  def __init__(self, executable_path, default_args=None):
    super(RemoteProcessRunner, self).__init__(
        executable_path, default_args=default_args)

  def run(self, **kwargs):  # pylint: disable=arguments-differ
    # TODO(ochang): This can be implemented, but isn't necessary yet.
    raise NotImplementedError

  def run_and_wait(self,
                   additional_args=None,
                   timeout=None,
                   terminate_before_kill=False,
                   terminate_wait_time=None,
                   input_data=None,
                   max_stdout_len=None,
                   extra_env=None,
                   stdout=subprocess.PIPE,
                   stderr=subprocess.STDOUT,
                   **popen_args):
    # pylint: disable=unused-argument
    # pylint: disable=arguments-differ
    """Remote version of new_process.ProcessRunner.run_and_wait."""
    assert stdout == subprocess.PIPE
    assert stderr == subprocess.STDOUT

    request = untrusted_runner_pb2.RunAndWaitRequest(
        executable_path=self.executable_path,
        timeout=timeout,
        terminate_before_kill=terminate_before_kill,
        terminate_wait_time=terminate_wait_time,
        input_data=input_data,
        max_stdout_len=max_stdout_len)

    request.default_args.extend(self.default_args)
    request.additional_args.extend(additional_args)

    if 'bufsize' in popen_args:
      request.popen_args.bufsize = popen_args['bufsize']

    if 'executable' in popen_args:
      request.popen_args.executable = popen_args['executable']

    if 'shell' in popen_args:
      request.popen_args.shell = popen_args['shell']

    if 'cwd' in popen_args:
      request.popen_args.cwd = popen_args['cwd']

    passed_env = popen_args.get('env', None)
    if passed_env is not None:
      request.popen_args.env_is_set = True
      # Filter the passed environment to prevent leaking sensitive environment
      # variables if the caller passes e.g. os.environ.copy().
      environment.set_environment_vars(request.popen_args.env, passed_env)

    response = host.stub().RunAndWait(request)
    return process_result_from_proto(response.result)


def terminate_stale_application_instances():
  """Terminate stale application instances."""
  host.stub().TerminateStaleApplicationInstances(
      untrusted_runner_pb2.TerminateStaleApplicationInstancesRequest())



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Classes for dealing with Performance Analysis."""

import collections
import datetime
import json
import operator

from clusterfuzz._internal.metrics import fuzzer_logs

# A tuple representing an analyzer for a single problem.
BasicAnalyzer = collections.namedtuple('Analyzer', 'checker weight description')

# Thresholds to be used in analyzers.
LOGGING_THRESHOLD = 0.5
LOGGING_RUNS_THRESHOLD = 5
NEW_UNITS_THRESHOLD = 1
SLOW_UNITS_THRESHOLD = 5
SPEED_THRESHOLD = 1000

# Limit on examples provided for an issue.
EXAMPLES_LIMIT = 5

# Values of weights to be used for performance analysis.
# TODO(mmoroz): https://github.com/google/oss-fuzz/issues/196.
# These values have been brute-forced during testing over logs of 115 fuzzers.
# This part will be removed or re-implemented during ongoing work on performance
# analysis. The new implementation will detect more issues and store analysis
# results in BigQuery.
WEIGHTS_FOR_LIBFUZZER = {
    'bad_instrumentation': 256.0,
    'coverage': 1.0,
    'crash': 4.0,
    'leak': 128.0,
    'logging': 1.0,
    'oom': 8.0,
    'slow_unit': 4.0,
    'speed': 1.0,
    'startup_crash': 256.0,
    'timeout': 64.0,
}


def generate_report(issues, fuzzer_name, job_type):
  """Generate a simple json representing performance analysis results.

  Args:
    issues: List of performance issues with corresponding scores and attributes.
    fuzzer_name: Fuzzer name.
    job_type: Job name.

  Returns:
    A json report for the given performance issues.
  """
  report_dict = {
      'fuzzer_name': fuzzer_name,
      'job_type': job_type,
      'issues': issues,
  }
  return json.dumps(report_dict, indent=2)


class PerformanceAnalyzer(object):
  """Performance Analyzer."""

  def __init__(self, analyzers):
    """Inits the PerformanceAnalyzer.

    Args:
      analyzers: List of BasicAnalyzer tuples to be used for analysis.
    """
    self._analyzers = analyzers

  @property
  def analyzers(self):
    return self._analyzers

  def analyze_stats(self, stats):
    """Process the given fuzzer stats.

    Args:
      stats: A list of fuzzer stats rows extracted from BigQuery.

    Returns:
      A list of scores corresponding to given analyzers, a list of percents
      showing how many fuzzer runs have been affected by each issue and a set
      of examples of that issue.
    """
    num_analyzers = len(self.analyzers)
    performance_scores = [0] * num_analyzers
    affected_runs_percent = [0] * num_analyzers
    examples = [set() for _ in range(num_analyzers)]

    if not stats:
      return performance_scores, affected_runs_percent, examples

    for row in stats:
      current_scores = self.analyze_single_stats_row(row)
      performance_scores = list(
          map(operator.add, performance_scores, current_scores))
      current_runs = [int(score > 0.0) for score in current_scores]
      affected_runs_percent = list(
          map(operator.add, affected_runs_percent, current_runs))

      # Add upto max |EXAMPLES_LIMIT| examples for each issue type.
      for i in range(num_analyzers):
        if current_scores[i] > 0.0 and len(examples[i]) < EXAMPLES_LIMIT:
          time_of_run = datetime.datetime.utcfromtimestamp(row['timestamp'])
          examples[i].add(fuzzer_logs.get_log_relative_path(time_of_run))

    # Data normalization. Divide the scores by the number of logs analyzed.
    runs_count = len(stats)
    performance_scores = [(score / runs_count) for score in performance_scores]

    # Data normalization. Convert affected runs into percents of all logs.
    affected_runs_percent = [
        ((count * 100.0) / runs_count) for count in affected_runs_percent
    ]

    return performance_scores, affected_runs_percent, examples

  def analyze_single_stats_row(self, row):
    """Analyze single stats row by given analyzers.

    Args:
      row: A dict containing stats data to be analysed.

    Returns:
      A list of scores corresponding to given analyzers.
    """
    scores = []
    for analyzer in self.analyzers:
      scores.append(analyzer.checker(row) * analyzer.weight)

    return scores

  def get_issues(self, scores, percents, examples):
    """Get list of issues and corresponding scores for given raw scores.

    Args:
      scores: Raw scores calculated with self.analyzers.
      percents: Percents of logs affected by each issue.
      examples: Log names showing this particular issue.

    Returns:
      A list of issues affecting the fuzzer with corresponding scores.
    """
    scores_sorted = sorted(
        ((value, index) for index, value in enumerate(scores)), reverse=True)

    detected_issues = []

    for value, index in scores_sorted:
      if value == 0.0:
        break

      detected_issues.append({
          'type': self.analyzers[index].description,
          'percent': round(percents[index], 2),
          'examples': sorted(examples[index]),
          'score': round(value, 2)
      })

    if not detected_issues:
      detected_issues.append({
          'type': 'none',
          'percent': 100.0,
          'examples': [],
          'score': 0.0
      })

    return detected_issues


class LibFuzzerPerformanceAnalyzer(PerformanceAnalyzer):
  """LibFuzzer specific Performance Analyzer using BigQuery data."""

  def __init__(self):
    """Inits the LibFuzzerPerformanceAnalyzer."""
    # Create a list of analyzers for libFuzzer-based fuzzers.
    analyzers = [
        BasicAnalyzer(self.analyzer_bad_instrumentation,
                      WEIGHTS_FOR_LIBFUZZER['bad_instrumentation'],
                      'bad_instrumentation'),
        BasicAnalyzer(self.analyzer_coverage, WEIGHTS_FOR_LIBFUZZER['coverage'],
                      'coverage'),
        BasicAnalyzer(self.analyzer_crash, WEIGHTS_FOR_LIBFUZZER['crash'],
                      'crash'),
        BasicAnalyzer(self.analyzer_leak, WEIGHTS_FOR_LIBFUZZER['leak'],
                      'leak'),
        BasicAnalyzer(self.analyzer_logging, WEIGHTS_FOR_LIBFUZZER['logging'],
                      'logging'),
        BasicAnalyzer(self.analyzer_oom, WEIGHTS_FOR_LIBFUZZER['oom'], 'oom'),
        BasicAnalyzer(self.analyzer_slow_unit,
                      WEIGHTS_FOR_LIBFUZZER['slow_unit'], 'slow_unit'),
        BasicAnalyzer(self.analyzer_speed, WEIGHTS_FOR_LIBFUZZER['speed'],
                      'speed'),
        BasicAnalyzer(self.analyzer_startup_crash,
                      WEIGHTS_FOR_LIBFUZZER['startup_crash'], 'startup_crash'),
        BasicAnalyzer(self.analyzer_timeout, WEIGHTS_FOR_LIBFUZZER['timeout'],
                      'timeout'),
    ]

    PerformanceAnalyzer.__init__(self, analyzers)

  def analyzer_bad_instrumentation(self, stats):
    """Check if there is bad instrumentation flag in the given stats.
    Returns a number in range [0.0, 1.0]."""
    if stats['bad_instrumentation']:
      return 1.0

    return 0.0

  def analyzer_coverage(self, stats):
    """Calculate a number of new units using the given stats data.
    Returns a number in range [0.0, 1.0]."""
    if stats.get('new_units_added') is None:
      # The stats row does not have new units info, i.e. there is another issue.
      return 0.0

    if stats['strategy_corpus_subset']:
      # We will always increase coverage when starting from a small subset, so
      # ignore these results.
      return 0.0

    if not stats['actual_duration']:
      return 0.0

    # Prorate the threshold in accordance with actual execution time.
    threshold = NEW_UNITS_THRESHOLD * (
        float(stats['actual_duration']) / stats['expected_duration'])
    if stats['new_units_added'] >= threshold:
      # The fuzzer finds NEW units quickly enough, coverage is not a problem.
      return 0.0

    # Result approaches 1.0 as fuzzer finds fewer new units.
    result = float(threshold - stats['new_units_added']) / threshold
    return min(1.0, result)

  def analyzer_crash(self, stats):
    """Check if there is a crash (ASan/MSan/UBSan) in the given stats.
    Returns a number in range [0.0, 1.0]."""
    if stats['crash_count']:
      return 1.0

    return 0.0

  def analyzer_leak(self, stats):
    """Check if there is a LeakSanitizer crash in the given stats.
    Returns a number in range [0.0, 1.0]."""
    if stats['leak_count']:
      return 1.0

    return 0.0

  def analyzer_logging(self, stats):
    """Calculate the fraction of redundant log lines of all logs lines.
    Returns number in range [0.0, 1.0]."""
    if stats['log_lines_from_engine'] < LOGGING_RUNS_THRESHOLD:
      # That means that fuzzer doesn't work and there is another problem, e.g.
      # a starup crash, or some troubles with configs, or anything else.
      return 0.0

    libfuzzer_and_other_lines_count = (
        stats['log_lines_from_engine'] + stats['log_lines_unwanted'])

    logging_score = (
        float(stats['log_lines_unwanted']) / libfuzzer_and_other_lines_count)

    if logging_score < LOGGING_THRESHOLD:
      # Ignore logging score if it's fewer than the threshold, as it often can
      # be greater that 0.0, but is not meaningful in terms of performance.
      logging_score = 0.0

    return logging_score

  def analyzer_oom(self, stats):
    """Check if there is a OOM crash in the given stats.
    Returns a number in range [0.0, 1.0]."""
    if stats['oom_count']:
      return 1.0

    return 0.0

  def analyzer_slow_unit(self, stats):
    """Calculates a number of slow units in the given stats.
    Returns a number in range [0.0, 1.0]."""
    if not stats['actual_duration']:
      return 0.0

    # Prorate the threshold in accordance with actual execution time.
    threshold = SLOW_UNITS_THRESHOLD * (
        float(stats['actual_duration']) / stats['expected_duration'])

    # A couple of slow units during fuzzing session is not that bad. This is
    # why we use a relative score below.
    return min(1.0, stats['slow_units_count'] / threshold)

  def analyzer_speed(self, stats):
    """Extract execution speed and compare to the threshold value.
    Returns a number in range [0.0, 1.0]."""
    if stats.get('average_exec_per_sec') is None:
      # The stats row does have speed info, i.e. there is another issue.
      return 0.0

    if stats['strategy_corpus_subset']:
      # When starting from a super small subset, we might have completely
      # different speed than when working from a large corpus, so ignore the
      # results.
      return 0.0

    if not stats['average_exec_per_sec']:
      # That means we have no stats in the log, there should be another problem.
      return 0.0

    # FIXME(mmoroz): needs more brainstorming. The same threshold value can be
    # unreachable for some fuzzers and too small for others.
    if stats['average_exec_per_sec'] >= SPEED_THRESHOLD:
      return 0.0

    # Result approaches 1.0 as fuzzer works slower.
    result = (
        float(SPEED_THRESHOLD - stats['average_exec_per_sec']) /
        SPEED_THRESHOLD)
    return min(1.0, result)

  def analyzer_startup_crash(self, stats):
    """Check if fuzzer had crashed on a startup."""
    if stats['startup_crash_count']:
      return 1.0

    return 0.0

  def analyzer_timeout(self, stats):
    """Check if there is a libFuzzer timeout in the stats.
    Returns a number in range [0.0, 1.0]."""
    if stats['timeout_count']:
      return 1.0

    return 0.0



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""big_query_query tests."""

import unittest

from libs.query import big_query_query


class FilterTest(unittest.TestCase):
  """Test filtering all types."""

  def test_single(self):
    """Test filtering single values."""
    q = big_query_query.Query()
    q.filter('int', 123)
    q.filter('float', 123.456)
    q.filter('bool', True)
    q.filter('string', 'test"t\nest')

    self.assertEqual(('(int = 123 AND float = 123.456 AND bool = true AND '
                      'string = "test\\"t\\nest")'), q.get_where_clause())

  def test_array(self):
    """Test filtering multiple values."""
    q = big_query_query.Query()
    q.filter_in('multi', [1, 1.2, True, False, 'test"test'])

    self.assertEqual('(multi IN (1, 1.2, true, false, "test\\"test"))',
                     q.get_where_clause())


class UnionTest(unittest.TestCase):
  """Test union."""

  def test_two_groups(self):
    """Test OR in two separate groups."""
    q = big_query_query.Query()
    q.filter('q', 'test')

    s1_1 = q.new_subquery()
    s1_1.filter('s1_1', 'test')
    s1_2 = q.new_subquery()
    s1_2.filter('s1_2', 'test')

    s2_1 = q.new_subquery()
    s2_1.filter('s2_1', 'test')
    s2_2 = q.new_subquery()
    s2_2.filter('s2_2', 'test')

    q.union(s1_1, s1_2)
    q.union(s2_1, s2_2)

    self.assertEqual(
        ('(q = "test" AND ((s1_1 = "test") OR (s1_2 = "test")) AND '
         '((s2_1 = "test") OR (s2_2 = "test")))'), q.get_where_clause())

  def test_nested_queries(self):
    """Test nested OR."""
    q = big_query_query.Query()
    q.filter('q', 'test')

    s1_1 = q.new_subquery()
    s1_1.filter('s1_1', 'test')
    s1_2 = q.new_subquery()
    s1_2.filter('s1_2', 'test')

    s2_1 = q.new_subquery()
    s2_1.filter('s2_1', 'test')
    s2_2 = q.new_subquery()
    s2_2.filter('s2_2', 'test')

    q.union(s1_1, s1_2)
    s1_2.union(s2_1, s2_2)

    self.assertEqual(('(q = "test" AND ((s1_1 = "test") OR (s1_2 = "test" AND '
                      '((s2_1 = "test") OR (s2_2 = "test")))))'),
                     q.get_where_clause())



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""py_unittest.py runs tests under src/appengine and butler/tests"""

import io
import itertools
import logging
import multiprocessing
import os
import platform
import signal
import sys
import time
import traceback
import unittest

from local.butler import appengine
from local.butler import common
from src.clusterfuzz._internal.config import local_config

APPENGINE_TEST_DIRECTORY = os.path.join('src', 'clusterfuzz', '_internal',
                                        'tests', 'appengine')
CORE_TEST_DIRECTORY = os.path.join('src', 'clusterfuzz', '_internal', 'tests',
                                   'core')
SLOW_TEST_THRESHOLD = 2  # In seconds.
TESTS_TIMEOUT = 20 * 60  # In seconds.


class TrackedTestResult(unittest.TextTestResult):
  """Result object that tracks slow-running tests."""

  def __init__(self, *args, **kwargs):
    super(TrackedTestResult, self).__init__(*args, **kwargs)
    self.slow_tests = []

  def startTest(self, test):
    self._start_time = time.time()
    super(TrackedTestResult, self).startTest(test)

  def addSuccess(self, test):
    elapsed_time = time.time() - self._start_time
    super(TrackedTestResult, self).addSuccess(test)

    if elapsed_time <= SLOW_TEST_THRESHOLD:
      return

    description = self.getDescription(test).splitlines()[0]
    self.slow_tests.append((elapsed_time, description))


class TrackedTestRunner(unittest.TextTestRunner):
  """TextTestRunner wrapper that reports additional information we collect."""

  def __init__(self, *args, **kwargs):
    kwargs['resultclass'] = TrackedTestResult
    super(TrackedTestRunner, self).__init__(*args, **kwargs)

  def run(self, test):
    result = super(TrackedTestRunner, self).run(test)

    if not result.slow_tests:
      return result

    self.stream.writeln('\nSlow tests:')
    for elapsed_time, test_name in sorted(result.slow_tests, reverse=True):
      print('%6.2fs: %s' % (elapsed_time, test_name))

    return result


class TestResult(object):
  """Test results."""

  def __init__(self, output, num_errors, num_failures, num_skipped, total_run):
    self.output = output
    self.num_errors = num_errors
    self.num_failures = num_failures
    self.num_skipped = num_skipped
    self.total_run = total_run


def test_worker_init():
  """Initialise test worker process."""
  if platform.system() != 'Windows':
    # Prevent KeyboardInterrupt error output.
    signal.signal(signal.SIGINT, signal.SIG_IGN)


def run_one_test_parallel(args):
  """Test worker."""
  try:
    os.environ['PARALLEL_TESTS'] = '1'

    test_modules, suppress_output = args
    suite = unittest.loader.TestLoader().loadTestsFromNames(test_modules)

    stream = io.StringIO()

    # Verbosity=0 since we cannot see real-time test execution order when tests
    # are executed in parallel.
    tests = ', '.join(test_modules)
    print('Running', tests)
    result = unittest.TextTestRunner(
        stream=stream, verbosity=0, buffer=suppress_output).run(suite)
    print('Done running', tests)

    stream.flush()
    value = stream.getvalue()

    return TestResult(value, len(result.errors), len(result.failures),
                      len(result.skipped), result.testsRun)
  except BaseException:
    # Print exception traceback here, as it will be lost otherwise.
    traceback.print_exc()
    raise


def run_tests_single_core(args, test_directory, top_level_dir):
  """Run tests (single CPU)."""
  suites = unittest.loader.TestLoader().discover(
      test_directory, pattern=args.pattern, top_level_dir=top_level_dir)

  # TODO(mbarbella): Re-implement code coverage after migrating to Python 3.
  # Verbosity=2 since we want to see real-time test execution with test name
  # and result.
  result = TrackedTestRunner(
      verbosity=2, buffer=(not args.unsuppress_output)).run(suites)

  if result.errors or result.failures:
    sys.exit(1)


def run_tests_parallel(args, test_directory, top_level_dir):
  """Run tests (multiple CPUs)."""
  suites = unittest.loader.TestLoader().discover(
      test_directory, pattern=args.pattern, top_level_dir=top_level_dir)

  test_classes = []  # pylint: disable=protected-access
  for suite in suites:
    for subsuite in suite._tests:  # pylint: disable=protected-access
      # According to:
      # https://github.com/python/cpython/blob/2.7/Lib/unittest/loader.py#L24,
      # this is how we can get a ModuleImportFailure error.
      if subsuite.__class__.__name__ == 'ModuleImportFailure':
        unittest.TextTestRunner(verbosity=1).run(subsuite)
        raise Exception('A failure occurred while importing the module.')

      for test_class in subsuite._tests:  # pylint: disable=protected-access
        test_classes.append((test_class.__module__,
                             test_class.__class__.__name__))
  test_classes = sorted(test_classes)

  test_modules = []
  for module_path, _ in itertools.groupby(test_classes, key=lambda k: k[0]):
    test_modules.append(module_path)
  test_modules = sorted(test_modules)

  cpu_count = multiprocessing.cpu_count()
  pool = multiprocessing.Pool(cpu_count, test_worker_init)

  total_result = TestResult('', 0, 0, 0, 0)

  # partition tests
  test_args = []

  tests_per_cpu = max(1, len(test_modules) // cpu_count)
  for i in range(0, len(test_modules), tests_per_cpu):
    group = test_modules[i:i + tests_per_cpu]
    test_args.append((group, not args.unsuppress_output))

  results = pool.map_async(run_one_test_parallel, test_args)

  while True:
    try:
      # KeyboardInterrupt never gets raised unless we pass a timeout.
      results = results.get(timeout=TESTS_TIMEOUT)
      break
    except KeyboardInterrupt:
      pool.terminate()
      pool.join()
      sys.exit(1)

  pool.close()
  pool.join()

  for result in results:
    if result.num_failures or result.num_errors:
      print(result.output)

    total_result.num_errors += result.num_errors
    total_result.num_failures += result.num_failures
    total_result.num_skipped += result.num_skipped
    total_result.total_run += result.total_run

  print('Ran %d tests (%d skipped, %d errors, %d failures).' %
        (total_result.total_run, total_result.num_skipped,
         total_result.num_errors, total_result.num_failures))

  if total_result.num_errors or total_result.num_failures:
    sys.exit(1)


def execute(args):
  """Run Python unit tests. For unittests involved appengine, sys.path needs
  certain modification."""
  os.environ['PY_UNITTESTS'] = 'True'

  if os.getenv('INTEGRATION') or os.getenv('UNTRUSTED_RUNNER_TESTS'):
    # Set up per-user buckets used by integration tests.
    os.environ['CORPUS_BUCKET'] = common.test_bucket('TEST_CORPUS_BUCKET')
    os.environ['QUARANTINE_BUCKET'] = common.test_bucket(
        'TEST_QUARANTINE_BUCKET')
    os.environ['BACKUP_BUCKET'] = common.test_bucket('TEST_BACKUP_BUCKET')
    os.environ['COVERAGE_BUCKET'] = common.test_bucket('TEST_COVERAGE_BUCKET')

  # Kill leftover instances of emulators and dev appserver.
  common.kill_leftover_emulators()

  # Don't use absolute paths to make it easier to compare results in tests.
  os.environ['CONFIG_DIR_OVERRIDE'] = os.path.join('.', 'configs', 'test')

  top_level_dir = os.path.join('src', 'clusterfuzz', '_internal')
  if args.target == 'appengine':
    # Build template files.
    appengine.build_templates()

    test_directory = APPENGINE_TEST_DIRECTORY
    sys.path.insert(0, os.path.abspath(os.path.join('src', 'appengine')))

    for i, path in enumerate(sys.path):
      if 'third_party' in path:
        # Replace third_party with App Engine third_party/.
        sys.path[i] = os.path.abspath(
            os.path.join('src', 'appengine', 'third_party'))

    if sys.version_info.major == 2:
      # TODO(ochang): Remove once migrated to Python 3.
      appengine_sdk_path = appengine.find_sdk_path()
      sys.path.insert(0, appengine_sdk_path)

      # Get additional App Engine third party imports.
      import dev_appserver
      dev_appserver.fix_google_path()
      sys.path.extend(dev_appserver.EXTRA_PATHS)

      # Loading appengine_main from the current project ensures that any
      # changes to configuration there are available to all tests (e.g.
      # sys.path modifications, namespaces, etc.)
      try:
        from src.appengine import main as appengine_main
        (appengine_main)  # pylint: disable=pointless-statement
      except ImportError:
        print('Note: unable to import appengine_main.')

      # google.auth uses App Engine credentials based on importability of
      # google.appengine.api.app_identity.
      try:
        from google.auth import app_engine as auth_app_engine
        if auth_app_engine.app_identity:
          auth_app_engine.app_identity = None
      except ImportError:
        pass
  elif args.target == 'core':
    test_directory = CORE_TEST_DIRECTORY
  else:
    # Config module tests.
    os.environ['CONFIG_DIR_OVERRIDE'] = args.config_dir
    test_directory = os.path.join(args.config_dir, 'modules')
    top_level_dir = None

    # Modules may use libs from our App Engine directory.
    sys.path.insert(0, os.path.abspath(os.path.join('src', 'appengine')))

    # Fix paths again to get config modules added to the import path.
    from clusterfuzz._internal.base import modules
    modules.fix_module_search_paths()

  # Set expected environment variables.
  local_config.ProjectConfig().set_environment()

  # Needed for NDB to work with cloud datastore emulator.
  os.environ['DATASTORE_USE_PROJECT_ID_AS_APP_ID'] = 'true'

  if args.verbose:
    # Force logging to console for this process and child processes.
    os.environ['LOG_TO_CONSOLE'] = 'True'
  else:
    # Disable logging.
    logging.disable(logging.CRITICAL)

  if args.pattern is None:
    args.pattern = '*_test.py'

  if args.parallel:
    # TODO(tanin): Support coverage.
    run_tests_parallel(args, test_directory, top_level_dir)
  else:
    run_tests_single_core(args, test_directory, top_level_dir)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for strategy selection file."""

import unittest

from google.cloud import ndb

from clusterfuzz._internal.bot.fuzzers import strategy_selection
from clusterfuzz._internal.bot.tasks import fuzz_task
from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.fuzzing import strategy
from clusterfuzz._internal.system import environment
from clusterfuzz._internal.tests.test_libs import helpers as test_helpers
from clusterfuzz._internal.tests.test_libs import test_utils


class TestDefaultStrategySelectionLibFuzzerPatched(unittest.TestCase):
  """Tests whether program properly generates strategy pools for use by the
  libFuzzer launcher."""

  def setUp(self):
    """Set up method for strategy pool generator tests with patch."""
    test_helpers.patch_environ(self)
    test_helpers.patch(self, [
        'clusterfuzz._internal.bot.fuzzers.engine_common.decide_with_probability'
    ])
    self.mock.decide_with_probability.return_value = True

  def test_default_pool_deterministic(self):
    """Deterministically tests the default strategy pool generator."""
    strategy_pool = strategy_selection.generate_default_strategy_pool(
        strategy_list=strategy.LIBFUZZER_STRATEGY_LIST, use_generator=True)

    # Ml rnn and radamsa strategies are mutually exclusive. Because of how we
    # patch, ml rnn will evaluate to false, however this depends on the
    # implementation.
    self.assertTrue(
        strategy_pool.do_strategy(strategy.CORPUS_MUTATION_RADAMSA_STRATEGY))
    self.assertFalse(
        strategy_pool.do_strategy(strategy.CORPUS_MUTATION_ML_RNN_STRATEGY))
    self.assertTrue(strategy_pool.do_strategy(strategy.CORPUS_SUBSET_STRATEGY))
    self.assertTrue(
        strategy_pool.do_strategy(strategy.RANDOM_MAX_LENGTH_STRATEGY))
    self.assertTrue(
        strategy_pool.do_strategy(strategy.RECOMMENDED_DICTIONARY_STRATEGY))
    self.assertTrue(strategy_pool.do_strategy(strategy.VALUE_PROFILE_STRATEGY))
    self.assertTrue(strategy_pool.do_strategy(strategy.FORK_STRATEGY))
    self.assertTrue(strategy_pool.do_strategy(strategy.MUTATOR_PLUGIN_STRATEGY))


class TestStrategySelectionLibFuzzerPatchless(unittest.TestCase):
  """Tests to see whether a default strategy pool is properly generated by the
  file for the libFuzzer launcher."""

  def test_default_pool_generator(self):
    """Ensures that a call to generate_default_strategy_pool does not yield an
    exception. Deterministic behaviors are tested in the previous test."""
    strategy_selection.generate_default_strategy_pool(
        strategy_list=strategy.LIBFUZZER_STRATEGY_LIST, use_generator=True)


@test_utils.with_cloud_emulators('datastore')
class TestMultiArmedBanditStrategySelectionLibFuzzerPatch(unittest.TestCase):
  """Tests whether a multi armed bandit strategy pool is properly
  generated according to the specified distribution for the libFuzzer
  launcher."""

  def setUp(self):
    """Put data in the local ndb table the tests to query from and set
    bandit selection environment variable."""
    test_helpers.patch_environ(self)

    data = []

    strategy1 = data_types.FuzzStrategyProbability()
    strategy1.strategy_name = 'fork,corpus_subset,recommended_dict,'
    strategy1.probability = 0.33
    strategy1.engine = 'libFuzzer'
    data.append(strategy1)

    strategy2 = data_types.FuzzStrategyProbability()
    strategy2.strategy_name = ('random_max_len,corpus_mutations_ml_rnn,'
                               'value_profile,recommended_dict,')
    strategy2.probability = 0.34
    strategy2.engine = 'libFuzzer'
    data.append(strategy2)

    strategy3 = data_types.FuzzStrategyProbability()
    strategy3.strategy_name = ('corpus_mutations_radamsa,'
                               'random_max_len,corpus_subset,')
    strategy3.probability = 0.33
    strategy3.engine = 'libFuzzer'
    data.append(strategy3)
    ndb.put_multi(data)

    distribution = fuzz_task.get_strategy_distribution_from_ndb()

    environment.set_value('USE_BANDIT_STRATEGY_SELECTION', True)
    environment.set_value('STRATEGY_SELECTION_DISTRIBUTION', distribution)

  def test_multi_armed_bandit_strategy_pool(self):
    """Ensures a call to the multi armed bandit strategy selection function
    doesn't yield an exception through any of the experimental paths."""
    environment.set_value('STRATEGY_SELECTION_METHOD', 'default')
    strategy_selection.generate_weighted_strategy_pool(
        strategy_list=strategy.LIBFUZZER_STRATEGY_LIST,
        use_generator=True,
        engine_name='libFuzzer')
    environment.set_value('STRATEGY_SELECTION_METHOD', 'multi_armed_bandit')
    strategy_selection.generate_weighted_strategy_pool(
        strategy_list=strategy.LIBFUZZER_STRATEGY_LIST,
        use_generator=True,
        engine_name='libFuzzer')


@test_utils.with_cloud_emulators('datastore')
class TestMultiArmedBanditStrategySelectionLibFuzzer(unittest.TestCase):
  """Tests whether multi armed bandit strategy pool is properly generated
  according to the specified distribution for the libFuzzer launcher.

  Deterministic tests. Only one strategy is put in the ndb table upon setup,
  so we know what the drawn strategy pool should be."""

  def setUp(self):
    """Put data in the local ndb table the tests to query from."""
    test_helpers.patch_environ(self)
    test_helpers.patch(self, [
        'clusterfuzz._internal.bot.fuzzers.engine_common.decide_with_probability'
    ])
    self.mock.decide_with_probability.return_value = True

    data = []

    strategy1 = data_types.FuzzStrategyProbability()
    strategy1.strategy_name = ('random_max_len,corpus_mutations_ml_rnn,'
                               'value_profile,recommended_dict,')
    strategy1.probability = 1
    strategy1.engine = 'libFuzzer'
    data.append(strategy1)
    ndb.put_multi(data)

    distribution = fuzz_task.get_strategy_distribution_from_ndb()

    environment.set_value('USE_BANDIT_STRATEGY_SELECTION', True)
    environment.set_value('STRATEGY_SELECTION_DISTRIBUTION', distribution)

  def test_weighted_strategy_pool(self):
    """Tests whether a proper strategy pool is returned by the multi armed
    bandit selection implementation with medium temperature.

    Based on deterministic strategy selection. Mutator plugin is patched to
    be included in our strategy pool."""
    environment.set_value('STRATEGY_SELECTION_METHOD', 'multi_armed_bandit')
    strategy_pool = strategy_selection.generate_weighted_strategy_pool(
        strategy_list=strategy.LIBFUZZER_STRATEGY_LIST,
        use_generator=True,
        engine_name='libFuzzer')
    self.assertTrue(
        strategy_pool.do_strategy(strategy.CORPUS_MUTATION_ML_RNN_STRATEGY))
    self.assertTrue(
        strategy_pool.do_strategy(strategy.RANDOM_MAX_LENGTH_STRATEGY))
    self.assertTrue(strategy_pool.do_strategy(strategy.VALUE_PROFILE_STRATEGY))
    self.assertTrue(
        strategy_pool.do_strategy(strategy.RECOMMENDED_DICTIONARY_STRATEGY))
    self.assertFalse(
        strategy_pool.do_strategy(strategy.CORPUS_MUTATION_RADAMSA_STRATEGY))
    self.assertFalse(strategy_pool.do_strategy(strategy.FORK_STRATEGY))


class TestDefaultStrategySelectionAFLPatched(unittest.TestCase):
  """Tests whether program properly generates strategy pools for use by the
  AFL launcher."""

  def setUp(self):
    """Set up method for strategy pool generator tests with patch."""
    test_helpers.patch_environ(self)
    test_helpers.patch(self, [
        'clusterfuzz._internal.bot.fuzzers.engine_common.decide_with_probability'
    ])
    self.mock.decide_with_probability.return_value = True

  def test_default_pool_deterministic(self):
    """Deterministically tests the default strategy pool generator."""
    strategy_pool = strategy_selection.generate_default_strategy_pool(
        strategy_list=strategy.AFL_STRATEGY_LIST, use_generator=True)

    # Ml rnn and radamsa strategies are mutually exclusive. Because of how we
    # patch, ml rnn will evaluate to false, however this depends on the
    # implementation.
    self.assertTrue(
        strategy_pool.do_strategy(strategy.CORPUS_MUTATION_RADAMSA_STRATEGY))
    self.assertFalse(
        strategy_pool.do_strategy(strategy.CORPUS_MUTATION_ML_RNN_STRATEGY))
    self.assertTrue(strategy_pool.do_strategy(strategy.CORPUS_SUBSET_STRATEGY))


class TestStrategySelectionAFLPatchless(unittest.TestCase):
  """Tests to see whether a default strategy pool is properly generated by the
  file for the AFL launcher."""

  def test_default_pool_generator(self):
    """Ensures that a call to generate_default_strategy_pool does not yield an
    exception. Deterministic behaviors are tested in the previous test."""
    strategy_selection.generate_default_strategy_pool(
        strategy_list=strategy.AFL_STRATEGY_LIST, use_generator=True)


@test_utils.with_cloud_emulators('datastore')
class TestMultiArmedBanditStrategySelectionAFLPatch(unittest.TestCase):
  """Tests whether a multi armed bandit strategy pool is properly
  generated according to the specified distribution for the AFL launcher."""

  def setUp(self):
    """Put data in the local ndb table the tests to query from and set
    bandit selection environment variable."""
    test_helpers.patch_environ(self)

    data = []

    strategy1 = data_types.FuzzStrategyProbability()
    strategy1.strategy_name = 'corpus_mutations_ml_rnn,corpus_subset,'
    strategy1.probability = 0.33
    strategy1.engine = 'afl'
    data.append(strategy1)

    strategy2 = data_types.FuzzStrategyProbability()
    strategy2.strategy_name = ('corpus_mutations_radamsa,corpus_subset,')
    strategy2.probability = 0.34
    strategy2.engine = 'afl'
    data.append(strategy2)

    strategy3 = data_types.FuzzStrategyProbability()
    strategy3.strategy_name = ('corpus_subset,')
    strategy3.probability = 0.33
    strategy3.engine = 'afl'
    data.append(strategy3)
    ndb.put_multi(data)

    distribution = fuzz_task.get_strategy_distribution_from_ndb()

    environment.set_value('USE_BANDIT_STRATEGY_SELECTION', True)
    environment.set_value('STRATEGY_SELECTION_DISTRIBUTION', distribution)

  def test_multi_armed_bandit_strategy_pool(self):
    """Ensures a call to the multi armed bandit strategy selection function
    doesn't yield an exception through any of the experimental paths."""
    environment.set_value('STRATEGY_SELECTION_METHOD', 'default')
    strategy_selection.generate_weighted_strategy_pool(
        strategy_list=strategy.AFL_STRATEGY_LIST,
        use_generator=True,
        engine_name='afl')
    environment.set_value('STRATEGY_SELECTION_METHOD', 'multi_armed_bandit')
    strategy_selection.generate_weighted_strategy_pool(
        strategy_list=strategy.AFL_STRATEGY_LIST,
        use_generator=True,
        engine_name='afl')


@test_utils.with_cloud_emulators('datastore')
class TestMultiArmedBanditStrategySelectionAFL(unittest.TestCase):
  """Tests whether multi armed bandit strategy pool is properly generated
  according to the specified distribution for the AFL launcher.

  Deterministic tests. Only one strategy is put in the ndb table upon setup,
  so we know what the drawn strategy pool should be."""

  def setUp(self):
    """Put data in the local ndb table the tests to query from."""
    test_helpers.patch_environ(self)
    test_helpers.patch(self, [
        'clusterfuzz._internal.bot.fuzzers.engine_common.decide_with_probability'
    ])
    self.mock.decide_with_probability.return_value = True

    data = []

    strategy1 = data_types.FuzzStrategyProbability()
    strategy1.strategy_name = 'corpus_mutations_ml_rnn,corpus_subset,'
    strategy1.probability = 1
    strategy1.engine = 'afl'
    data.append(strategy1)
    ndb.put_multi(data)

    distribution = fuzz_task.get_strategy_distribution_from_ndb()

    environment.set_value('USE_BANDIT_STRATEGY_SELECTION', True)
    environment.set_value('STRATEGY_SELECTION_DISTRIBUTION', distribution)

  def test_weighted_strategy_pool(self):
    """Tests whether a proper strategy pool is returned by the multi armed
    bandit selection implementation with medium temperature.

    Based on deterministic strategy selection. Mutator plugin is patched to
    be included in our strategy pool."""
    environment.set_value('STRATEGY_SELECTION_METHOD', 'multi_armed_bandit')
    strategy_pool = strategy_selection.generate_weighted_strategy_pool(
        strategy_list=strategy.AFL_STRATEGY_LIST,
        use_generator=True,
        engine_name='afl')
    self.assertTrue(
        strategy_pool.do_strategy(strategy.CORPUS_MUTATION_ML_RNN_STRATEGY))
    self.assertFalse(
        strategy_pool.do_strategy(strategy.CORPUS_MUTATION_RADAMSA_STRATEGY))
    self.assertTrue(strategy_pool.do_strategy(strategy.CORPUS_SUBSET_STRATEGY))



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Build setup (untrusted side)."""

from clusterfuzz._internal.build_management import build_manager
from clusterfuzz._internal.protos import untrusted_runner_pb2
from clusterfuzz._internal.system import environment


def _build_response(result):
  if not result:
    return untrusted_runner_pb2.SetupBuildResponse(result=False)

  return untrusted_runner_pb2.SetupBuildResponse(
      result=True,
      app_path=environment.get_value('APP_PATH'),
      app_path_debug=environment.get_value('APP_PATH_DEBUG'),
      app_dir=environment.get_value('APP_DIR'),
      build_dir=environment.get_value('BUILD_DIR'),
      build_url=environment.get_value('BUILD_URL'),
      fuzz_target=environment.get_value('FUZZ_TARGET'),
      fuzz_target_count=environment.get_value('FUZZ_TARGET_COUNT'))


def setup_regular_build(request):
  """Set up a regular build."""
  build = build_manager.RegularBuild(request.base_build_dir, request.revision,
                                     request.build_url, request.target_weights,
                                     request.build_prefix)
  return _build_response(build.setup())



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Remote process implementation."""

from clusterfuzz._internal.metrics import logs
from clusterfuzz._internal.protos import untrusted_runner_pb2
from clusterfuzz._internal.system import new_process
from clusterfuzz._internal.system import process_handler

from . import protobuf_utils


def process_result_to_proto(process_result):
  """Convert new_process.ProcessResult to proto."""
  process_result_proto = untrusted_runner_pb2.ProcessResult(
      return_code=process_result.return_code,
      output=process_result.output,
      time_executed=process_result.time_executed,
      timed_out=process_result.timed_out)

  process_result_proto.command.extend(process_result.command)

  return process_result_proto


def run_and_wait(request, _):
  """Implementation of RunAndWait."""
  process_runner = new_process.ProcessRunner(request.executable_path,
                                             request.default_args)
  args = {}
  protobuf_utils.get_protobuf_field(args, request.popen_args, 'bufsize')
  protobuf_utils.get_protobuf_field(args, request.popen_args, 'executable')
  protobuf_utils.get_protobuf_field(args, request.popen_args, 'shell')
  protobuf_utils.get_protobuf_field(args, request.popen_args, 'cwd')

  if request.popen_args.env_is_set:
    args['env'] = request.popen_args.env
  else:
    args['env'] = None

  args['additional_args'] = request.additional_args
  protobuf_utils.get_protobuf_field(args, request, 'timeout')
  protobuf_utils.get_protobuf_field(args, request, 'terminate_before_kill')
  protobuf_utils.get_protobuf_field(args, request, 'terminate_wait_time')
  protobuf_utils.get_protobuf_field(args, request, 'input_data')
  protobuf_utils.get_protobuf_field(args, request, 'max_stdout_len')

  logs.log('Running command: %s' % process_runner.get_command())

  return untrusted_runner_pb2.RunAndWaitResponse(
      result=process_result_to_proto(process_runner.run_and_wait(**args)))


def run_process(request, _):
  """Implementation of RunProcess."""
  args = {}
  protobuf_utils.get_protobuf_field(args, request, 'cmdline')
  protobuf_utils.get_protobuf_field(args, request, 'current_working_directory')
  protobuf_utils.get_protobuf_field(args, request, 'timeout')
  protobuf_utils.get_protobuf_field(args, request, 'need_shell')

  if request.gestures:
    args['gestures'] = request.gestures

  if request.env_copy:
    args['env_copy'] = request.env_copy

  protobuf_utils.get_protobuf_field(args, request, 'testcase_run')
  protobuf_utils.get_protobuf_field(args, request, 'ignore_children')

  return_code, execution_time, output = process_handler.run_process(**args)
  response = untrusted_runner_pb2.RunProcessResponse(
      return_code=return_code, execution_time=execution_time, output=output)

  return response



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""External user permission utilities."""

from clusterfuzz._internal.base import memoize
from clusterfuzz._internal.base import utils
from clusterfuzz._internal.datastore import data_handler
from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.datastore import fuzz_target_utils
from clusterfuzz._internal.datastore import ndb_utils

MEMCACHE_TTL_IN_SECONDS = 15 * 60


def _fuzzers_for_job(job_type, include_parents):
  """Return all fuzzers that have the job associated.

  Args:
    job_type: The job type.
    include_parents: Include the parent fuzzer.

  Returns:
    A list of fuzzer names.
  """
  fuzzers = []
  engine_fuzzers = data_handler.get_fuzzing_engines()

  for fuzzer in data_types.Fuzzer.query(data_types.Fuzzer.jobs == job_type):
    # Add this if we're including all parents or this is not an engine fuzzer
    # with fuzz targets.
    if include_parents or fuzzer.name not in engine_fuzzers:
      fuzzers.append(fuzzer.name)

  for target_job in fuzz_target_utils.get_fuzz_target_jobs(job=job_type):
    fuzzers.append(target_job.fuzz_target_name)

  return sorted(fuzzers)


def _expand_prefix(all_names, prefix):
  """Expand the given prefix into real entity names.

  Args:
    all_names: A list of all entity names.
    prefix: A prefix string.

  Returns:
    A list of entity names that the pattern expands to.
  """
  return [name for name in all_names if name.startswith(prefix)]


def _get_permissions_query_for_user(user_email, entity_kind=None):
  """Get a permissions query for a given user.

  Args:
    user_email: The email of the user.
    entity_kind: The type (data_types.PermissionEntityKind) of the permission to
        filter by, or None.

  Returns:
    A ndb.Query giving the permissions for the given parameters.
  """
  permissions_for_user = data_types.ExternalUserPermission.query(
      data_types.ExternalUserPermission.email == utils.normalize_email(
          user_email))

  if entity_kind is not None:
    permissions_for_user = permissions_for_user.filter(
        data_types.ExternalUserPermission.entity_kind == entity_kind)

  return permissions_for_user


def _allowed_entities_for_user(user_email, entity_kind):
  """Return the entity names that the given user can access.

  Args:
    user_email: The email of the user.
    entity_kind: The type (data_types.PermissionEntityKind) of the entity.

  Returns:
    A list of entity names that the user has access to.
  """
  if not user_email:
    return []

  allowed = []
  permissions = _get_permissions_query_for_user(user_email, entity_kind)

  if entity_kind == data_types.PermissionEntityKind.FUZZER:
    all_names = data_handler.get_all_fuzzer_names_including_children()
  else:
    all_names = data_handler.get_all_job_type_names()

  for permission in permissions:
    if permission.is_prefix:
      allowed.extend(_expand_prefix(all_names, permission.entity_name))
    elif permission.entity_name in all_names:
      allowed.append(permission.entity_name)

  return sorted(allowed)


def _is_entity_allowed_for_user(user_email, name, entity_kind):
  """Return whether if the given user has access to the entity.

  Args:
    user_email: The email of the user.
    name: The name of the entity.
    entity_kind: The type of the entity.

  Returns:
    A bool indicating whether the given user has access to the entity.
  """
  if not user_email or not name:
    return False

  permissions = _get_permissions_query_for_user(user_email, entity_kind)

  for permission in permissions:
    if permission.is_prefix:
      if name.startswith(permission.entity_name):
        return True
    elif permission.entity_name == name:
      return True

  return False


def _allowed_users_for_entity(name, entity_kind, auto_cc=None):
  """Return a list of users that have permissions for the given entity.

  Args:
    name: The name of the entity.
    entity_kind: The type (data_types.PermissionEntityKind) of the entity.
    auto_cc: The Auto CC type (data_types.AutoCCType) to filter on, or None.

  Returns:
    A list of user emails that have permission to access the given entity.
  """
  if not name:
    return []

  # Easy case: direct matches.
  direct_match_permissions = data_types.ExternalUserPermission.query(
      data_types.ExternalUserPermission.entity_kind == entity_kind,
      data_types.ExternalUserPermission.entity_name == name,
      ndb_utils.is_false(data_types.ExternalUserPermission.is_prefix),
      projection=[data_types.ExternalUserPermission.email])
  if auto_cc is not None:
    direct_match_permissions = direct_match_permissions.filter(
        data_types.ExternalUserPermission.auto_cc == auto_cc)

  allowed_users = [permission.email for permission in direct_match_permissions]

  # Find all permissions where the prefix matches the fuzzer_name.
  # Unfortunately, Datastore doesn't give us an easy way of doing so. To iterate
  # through a smaller set than every single permission, get all permissions that
  # contain a prefix string <= than the actual fuzzer name and >= the first
  # character.
  prefix_match_permissions = data_types.ExternalUserPermission.query(
      data_types.ExternalUserPermission.entity_kind == entity_kind,
      data_types.ExternalUserPermission.entity_name <= name,
      data_types.ExternalUserPermission.entity_name >= name[0],
      ndb_utils.is_true(data_types.ExternalUserPermission.is_prefix),
      projection=[
          data_types.ExternalUserPermission.email,
          data_types.ExternalUserPermission.entity_name
      ])
  if auto_cc is not None:
    prefix_match_permissions = prefix_match_permissions.filter(
        data_types.ExternalUserPermission.auto_cc == auto_cc)

  for permission in prefix_match_permissions:
    if not permission.entity_name:
      # No external user should have an empty prefix (access to all
      # fuzzers/jobs).
      continue

    if name.startswith(permission.entity_name):
      allowed_users.append(permission.email)

  return sorted(allowed_users)


def _cc_users_for_entity(name, entity_type, security_flag):
  """Return CC users for entity."""
  users = _allowed_users_for_entity(name, entity_type,
                                    data_types.AutoCCType.ALL)

  if security_flag:
    users.extend(
        _allowed_users_for_entity(name, entity_type,
                                  data_types.AutoCCType.SECURITY))

  return sorted(users)


@memoize.wrap(memoize.Memcache(MEMCACHE_TTL_IN_SECONDS))
def allowed_fuzzers_for_user(user_email,
                             include_from_jobs=False,
                             include_parents=False):
  """Return allowed fuzzers for the given user.

  Args:
    user_email: The email of the user.
    include_from_jobs: Include all fuzzers for the allowed jobs of the user.
    include_parents: Include parent fuzzers when there is no explicit permission
        for the parent fuzzer, but there are permissions for its children as a
        result of the user's job permissions. Only applies when
        include_from_jobs is set.

  Returns:
    A list of fuzzer names for which this user is allowed to view information
    about.
  """
  allowed_fuzzers = _allowed_entities_for_user(
      user_email, data_types.PermissionEntityKind.FUZZER)

  if include_from_jobs:
    allowed_jobs = allowed_jobs_for_user(user_email)
    for allowed_job in allowed_jobs:
      allowed_fuzzers.extend(_fuzzers_for_job(allowed_job, include_parents))

    allowed_fuzzers = list(set(allowed_fuzzers))

  return sorted(allowed_fuzzers)


@memoize.wrap(memoize.Memcache(MEMCACHE_TTL_IN_SECONDS))
def allowed_jobs_for_user(user_email):
  """Return allowed jobs for the given user.

  Args:
    user_email: The email of the user.

  Returns:
    A list of job names for which this user is allowed to view information
    about.
  """
  return _allowed_entities_for_user(user_email,
                                    data_types.PermissionEntityKind.JOB)


def allowed_users_for_fuzzer(fuzzer_name):
  """Return allowed external users for the given fuzzer.

  Args:
    fuzzer_name: The name of the fuzzer.

  Returns:
    A list of user emails that are allowed to view information relating to this
    fuzzer.
  """
  # TODO(ochang): Once we support jobs, take that into account.
  return _allowed_users_for_entity(fuzzer_name,
                                   data_types.PermissionEntityKind.FUZZER)


def cc_users_for_fuzzer(fuzzer_name, security_flag):
  """Return external users that should be CC'ed according to the given rule.

  Args:
    fuzzer_name: The name of the fuzzer.
    security_flag: Whether or not the CC is for a security issue.

  Returns:
    A list of user emails that should be CC'ed.
  """
  return _cc_users_for_entity(
      fuzzer_name, data_types.PermissionEntityKind.FUZZER, security_flag)


def is_fuzzer_allowed_for_user(user_email, fuzzer_name,
                               include_from_jobs=False):
  """Return whether if the given user has access to the fuzzer.

  Args:
    user_email: The email of the user.
    fuzzer_name: The name of the fuzzer.
    include_from_jobs: Include all fuzzers for the allowed jobs of the user.

  Returns:
    A bool indicating whether the given user has access to the fuzzer.
  """
  is_allowed = _is_entity_allowed_for_user(
      user_email, fuzzer_name, data_types.PermissionEntityKind.FUZZER)

  if not is_allowed and include_from_jobs:
    is_allowed = fuzzer_name in allowed_fuzzers_for_user(
        user_email, include_from_jobs=True)

  return is_allowed


def is_job_allowed_for_user(user_email, job_type):
  """Return whether if the given user has access to the job.

  Args:
    user_email: The email of the user.
    job_type: The name of the job.

  Returns:
    A bool indicating whether the given user has access to the job.
  """
  return _is_entity_allowed_for_user(user_email, job_type,
                                     data_types.PermissionEntityKind.JOB)


def is_upload_allowed_for_user(user_email):
  """Return whether if the given user has upload permissions.

  Args:
    user_email: The email of the user.

  Returns:
    A bool indicating whether the given user has upload permissions.
  """
  permissions = _get_permissions_query_for_user(
      user_email, data_types.PermissionEntityKind.UPLOADER)
  return bool(permissions.get())


def cc_users_for_job(job_type, security_flag):
  """Return external users that should be CC'ed according to the given rule.

  Args:
    job_type: The name of the job
    security_flag: Whether or not the CC is for a security issue.

  Returns:
    A list of user emails that should be CC'ed.
  """
  return _cc_users_for_entity(job_type, data_types.PermissionEntityKind.JOB,
                              security_flag)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Pubsub tests."""

import time
import unittest

import six

from clusterfuzz._internal.google_cloud_utils import pubsub
from clusterfuzz._internal.tests.test_libs import helpers
from clusterfuzz._internal.tests.test_libs import test_utils

ACK_DEADLINE = 3

# Additional amount of time to sleep to let ack deadline elapse.
ACK_DEADLINE_WINDOW = 1
PROJECT_NAME = 'fake-project'


@test_utils.integration
@test_utils.with_cloud_emulators('pubsub')
class PubSubTest(unittest.TestCase):
  """Tests for pubsub."""

  def setUp(self):
    helpers.patch_environ(self)
    self.topic = pubsub.topic_name(PROJECT_NAME, 'test-topic')
    self.subscription = pubsub.subscription_name(PROJECT_NAME, 'subscription')

    self.client = pubsub.PubSubClient()
    self.client.create_topic(self.topic)
    self.client.create_subscription(
        self.subscription, self.topic, ack_deadline=ACK_DEADLINE)

  def test_pull_from_subscription(self):
    """Test pull_from_subscription."""
    self.client.publish(
        self.topic, [
            pubsub.Message(data=b'123'),
            pubsub.Message(data=b'123', attributes={'a': '1337'}),
            pubsub.Message(data=b'456'),
            pubsub.Message(data=b'456'),
        ])

    messages = self.client.pull_from_subscription(
        self.subscription, max_messages=1)
    self.assertEqual(1, len(messages))
    self.assertEqual(b'123', messages[0].data)
    self.assertIsNone(messages[0].attributes)

    messages = self.client.pull_from_subscription(
        self.subscription, max_messages=1)
    self.assertEqual(1, len(messages))
    self.assertEqual(b'123', messages[0].data)
    self.assertDictEqual({'a': '1337'}, messages[0].attributes)

    messages = self.client.pull_from_subscription(
        self.subscription, acknowledge=True)
    self.assertEqual(2, len(messages))

    for message in messages:
      self.assertEqual(b'456', message.data)

    # Test messages which were not acked in time. They will be re-sent and can
    # be re-pulled.
    time.sleep(ACK_DEADLINE + ACK_DEADLINE_WINDOW)
    messages = self.client.pull_from_subscription(self.subscription)
    self.assertEqual(2, len(messages))
    six.assertCountEqual(self, [
        {
            'data': 'MTIz',
        },
        {
            'data': 'MTIz',
            'attributes': {
                'a': '1337',
            }
        },
    ], [pubsub._message_to_dict(message) for message in messages])  # pylint: disable=protected-access

  def test_ack(self):
    """Test a single message ack."""
    self.client.publish(
        self.topic, [
            pubsub.Message(data=b'123'),
        ])

    messages = self.client.pull_from_subscription(
        self.subscription, max_messages=1)
    self.assertEqual(1, len(messages))

    # Acknowledging the message means it shouldn't get pulled again.
    messages[0].ack()
    time.sleep(ACK_DEADLINE + ACK_DEADLINE_WINDOW)

    messages = self.client.pull_from_subscription(
        self.subscription, max_messages=1)
    self.assertEqual(0, len(messages))

  def test_modify_ack_deadline(self):
    """Test modify ACK deadline."""
    self.client.publish(
        self.topic, [
            pubsub.Message(data=b'123'),
        ])
    messages = self.client.pull_from_subscription(
        self.subscription, max_messages=1)
    self.assertEqual(1, len(messages))

    # Make message instantly available again.
    messages[0].modify_ack_deadline(0)
    messages = self.client.pull_from_subscription(
        self.subscription, max_messages=1)
    self.assertEqual(1, len(messages))

  def test_list_topics(self):
    """Test listing topics."""
    expected = []
    for i in range(5):
      topic = pubsub.topic_name(PROJECT_NAME, 'topic-{}'.format(i))
      expected.append(topic)
      self.client.create_topic(topic)

    expected.append('projects/fake-project/topics/test-topic')

    topics = list(self.client.list_topics('projects/' + PROJECT_NAME))
    six.assertCountEqual(self, expected, topics)

    # Note: Page size appears to be ignored by the emulator. Even when creating
    # large amounts of topics to force paging, the nextPageToken returned is
    # buggy and results in infinite loops.
    topics = list(
        self.client.list_topics('projects/' + PROJECT_NAME, page_size=1))
    six.assertCountEqual(self, expected, topics)

  def test_list_topic_subscriptions(self):
    """Test listing topic subscriptions."""
    expected = []
    for i in range(5):
      subscription = pubsub.subscription_name(PROJECT_NAME, 'sub-{}'.format(i))
      expected.append(subscription)
      self.client.create_subscription(subscription, self.topic)

    expected.append('projects/fake-project/subscriptions/subscription')

    subscriptions = list(self.client.list_topic_subscriptions(self.topic))
    six.assertCountEqual(self, expected, subscriptions)

    # Note: Page size appears to be ignored by the emulator. Even when creating
    # large amounts of topics to force paging, the nextPageToken returned is
    # buggy and results in infinite loops.
    subscriptions = list(
        self.client.list_topic_subscriptions(self.topic, page_size=1))
    six.assertCountEqual(self, expected, subscriptions)

  def test_get_topic(self):
    """Test getting a topic."""
    topic = self.client.get_topic(self.topic)
    self.assertEqual(self.topic, topic.name)
    self.assertIsNone(topic.labels)

  def test_get_subscription(self):
    """Test getting a subscription."""
    subscription = self.client.get_subscription(self.subscription)
    self.assertEqual(self.subscription, subscription.name)
    self.assertEqual(self.topic, subscription.topic)
    self.assertEqual(ACK_DEADLINE, subscription.ack_deadline)
    self.assertIsNone(subscription.labels)

  def test_delete_topic(self):
    """Test deleting a topic."""
    self.client.delete_topic(self.topic)
    self.assertIsNone(self.client.get_topic(self.topic))

  def test_delete_subscription(self):
    """Test deleting a subscription."""
    self.client.delete_subscription(self.subscription)
    self.assertIsNone(self.client.get_subscription(self.subscription))



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Handler for the regression page."""

import json

from flask import request

from clusterfuzz._internal.google_cloud_utils import big_query
from handlers import base_handler
from libs import crash_access
from libs import filters
from libs import handler
from libs import helpers
from libs.query import big_query_query

PAGE_SIZE = 30

SQL = """
WITH
  # Deduplicate rows for the same testcase.
  unique_rows AS (
    SELECT
      testcase_id,
      ARRAY_AGG(crash_type)[OFFSET(0)] AS crash_type,
      ARRAY_AGG(crash_state)[OFFSET(0)] AS crash_state,
      ARRAY_AGG(security_flag)[OFFSET(0)] AS security_flag,
      ARRAY_AGG({prefix}_range_start)[OFFSET(0)] AS {prefix}_range_start,
      ARRAY_AGG({prefix}_range_end)[OFFSET(0)] AS {prefix}_range_end,
      ARRAY_AGG(created_at)[OFFSET(0)] AS created_at
    FROM main.{table_id}
    WHERE {where_clause}
    GROUP BY testcase_id
    ORDER BY created_at DESC
  )

SELECT
  crash_type, crash_state, security_flag,
  COUNT(testcase_id) AS count,
  ARRAY_AGG(STRUCT(
    testcase_id,
    {prefix}_range_start AS range_start,
    {prefix}_range_end AS range_end,
    created_at
  )) AS testcases
FROM unique_rows
GROUP BY crash_type, crash_state, security_flag
ORDER BY count DESC
"""


class RevisionFilter(filters.Filter):
  """Filter for revision."""

  def add(self, query, params):
    """Set query according to revision and type params."""
    if not params.get('revision'):
      raise helpers.EarlyExitException('Please specify the revision.', 400)

    prefix = params['type']
    revision = helpers.cast(params['revision'], int,
                            "'revision' must be an integer")
    query.raw_filter('{prefix}_range_start < {revision} AND '
                     '{revision} <= {prefix}_range_end'.format(
                         revision=json.dumps(revision), prefix=prefix))


class KeywordFilter(filters.Filter):
  """Filter for keyword."""

  def add(self, query, params):
    """Set query according to search param."""
    value = params.get('q', '')
    if filters.is_empty(value):
      return

    for keyword in value.split(' '):
      query.raw_filter(
          '(LOWER(crash_state) LIKE %s OR LOWER(crash_type) LIKE %s)' %
          (json.dumps('%%%s%%' % keyword.lower()),
           json.dumps('%%%s%%' % keyword.lower())))


class IncludeZeroFilter(filters.Filter):
  """Filter for including a range starting with 0."""

  def add(self, query, params):
    """Set query based on if range should include 0 or not."""
    value = params.get('includeZero', False)
    prefix = params['type']

    if not value:
      query.raw_filter('{prefix}_range_start > 0'.format(prefix=prefix))


class TypeFilter(filters.Filter):
  """Filter for type."""

  def add(self, query, params):  # pylint: disable=unused-argument
    """Validate type param is either regression or fixed."""
    value = params.get('type', 'regression')

    if value not in ['fixed', 'regression']:
      raise helpers.EarlyExitException(
          "'type' can only be either 'fixed' or 'regression'.", 400)


FILTERS = [
    TypeFilter(),
    RevisionFilter(),
    KeywordFilter(),
    IncludeZeroFilter(),
]


def get(params, query, offset, limit):
  """Get the data from BigQuery."""
  sql = SQL.format(
      table_id='%ss' % params['type'],
      where_clause=query.get_where_clause(),
      prefix=params['type'],
      offset=offset,
      limit=limit)
  client = big_query.Client()
  result = client.query(query=sql, offset=offset, limit=limit)
  return result.rows, result.total_count


def get_result():
  """Get the result for the crash stats page."""
  params = dict(request.iterparams())
  params['type'] = params.get('type', 'regression')
  page = helpers.cast(request.get('page') or 1, int, "'page' is not an int.")

  is_revision_empty = 'revision' not in params

  query = big_query_query.Query()
  crash_access.add_scope(query, params, 'security_flag', 'job_type',
                         'fuzzer_name')

  if is_revision_empty:
    total_count = 0
    rows = []
  else:
    filters.add(query, params, FILTERS)
    rows, total_count = get(
        params=params,
        query=query,
        offset=(page - 1) * PAGE_SIZE,
        limit=PAGE_SIZE)
    helpers.log('Regression', helpers.VIEW_OPERATION)

  result = {
      'totalPages': (total_count // PAGE_SIZE) + 1,
      'page': page,
      'pageSize': PAGE_SIZE,
      'items': rows,
      'totalCount': total_count,
      'isRevisionEmpty': is_revision_empty
  }
  return result, params


class Handler(base_handler.Handler):
  """Handler that lists testcases whose regression range contains a revision."""

  @handler.unsupported_on_local_server
  @handler.get(handler.HTML)
  def get(self):
    """Get and render the commit range in HTML."""
    result, params = get_result()
    return self.render('commit_range.html', {
        'result': result,
        'params': params
    })


class JsonHandler(base_handler.Handler):
  """JSON handler used for dynamic updates of commit ranges."""

  # See: https://bugs.chromium.org/p/chromium/issues/detail?id=760669
  @handler.post(handler.JSON, handler.JSON)
  @handler.oauth
  @handler.allowed_cors
  def post(self):
    """Get and render the commit range in JSON."""
    result, params = get_result()
    result['params'] = params
    return self.render_json(result)

  @handler.allowed_cors
  def options(self):
    """Responds with CORS headers."""
    return ''
